#! /usr/bin/env python3
# 
# Merges two versions of the same file, using an earlier version as a baseline
# (i.e., 3-way merge). If the baseline version is not supplied it is determined
# as the most recent backup older than both versions.
#
# Notes:
# - The merge process can easily get confused is there are parallel changes to
#   the same sections.
# - Use a visual merge script to help with manual reconciliation (e.g., via kdiff3).
#
# TODO:
# - Integrate merge heuristics for common issues.
#   -- For example, accounting for changes flagged with comments:
#      ## OLD: from regex import my_re
#      from my_regex import my_re
#   -- Accounting for changes elsewhere.
#

"""Perform 3-way merge with baseline from backup dir

Sample usage:

{prog} --stdout .bashrc ~/temp/.bashrc > .new-bashrc
"""

# Standard packages
import re

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla import system

# Constants for switches omitting leading dashes (e.g., DEBUG_MODE = "debug-mode")
BASENAME = "basename"
BACKUP_DIR = "backup-dir"
BACKUP_FILENAME = "backup-filename"
MAIN_FILENAME = "main-filename"
OTHER_FILENAME = "other-filename"
QUIET = "quiet"
USE_STDOUT = "stdout"
SKIP_BASELINE = "skip-baseline"
UPDATE_MAIN_FILE = "update-main-file"
IGNORE_ERRORS = "ignore-errors"
## OLD: MERGE = system.getenv_text("MERGE", "/usr/bin/merge -p",
MERGE = system.getenv_text("MERGE", "merge -p",
                           description="Command line for merge with output piping arg (e.g., -p)")
IGNORE_TIMESTAMP = system.getenv_bool("IGNORE_TIMESTAMP", False,
                                      "Whether to ignore time of backup files")

#...............................................................................

def get_timestamp(filename):
    """Returns timestamp for FILENAME as string value"""
    # EX: get_timestamp("/vmlinuz") => "2021-04-15 04:24:54"
    result = system.get_file_modification_time(filename, as_float=False)
    debug.trace(6, f"get_timestamp({filename}) => {result}")
    debug.assertion(isinstance(result, str))
    return result


def get_numeric_timestamp(filename):
    """Returns timestamp for FILENAME as floating point value"""
    # EX: get_numeric_timestamp("/vmlinuz") => 1618478694.0
    result = system.get_file_modification_time(filename, as_float=True)
    debug.trace(6, f"get_numeric_timestamp({filename}) => {result}")
    debug.assertion(isinstance(result, float))
    return result

#...............................................................................

class Script(Main):
    """Adhoc script processing class"""
    main_filename = ""
    other_filename = ""
    backup_filename = ""
    backup_dir = "./backup"
    basename = ""
    update_file1 = False
    skip_baseline = False
    quiet = False
    use_stdout = None
    ignore_errors = False
    ## TODO: use_temp_base_dir = True

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)
        self.main_filename = self.get_parsed_argument(MAIN_FILENAME)
        self.other_filename = self.get_parsed_argument(OTHER_FILENAME)
        self.backup_filename = self.get_parsed_option(BACKUP_FILENAME)
        default_backup_dir = gh.form_path(gh.dirname(self.main_filename),
                                          self.backup_dir) 
        self.backup_dir = self.get_parsed_option(BACKUP_DIR,
                                                 default_backup_dir)
        self.basename = self.get_parsed_option(BASENAME,
                                               gh.basename(self.main_filename))
        self.update_file1 = self.get_parsed_option(UPDATE_MAIN_FILE, self.update_file1)
        self.skip_baseline = self.get_parsed_option(SKIP_BASELINE, self.skip_baseline)
        self.quiet = self.get_parsed_option(QUIET, self.quiet)
        self.use_stdout = self.get_parsed_option(USE_STDOUT, not self.update_file1)
        self.ignore_errors = self.get_parsed_option(IGNORE_ERRORS)
        debug.trace_object(5, self, label="Script instance")

    def check_regular_file(self, filename):
        """Make sure regular file exists; otherwise, issue error and exit"""
        debug.trace(4, f"Script.check_regular_file({filename})")
        ok = (system.file_exists(filename) and system.is_regular_file(filename))
        if not ok and not self.ignore_errors:
            system.exit(f"Error: expecting regular file for '{filename}'")
        return ok

    def run_main_step(self):
        """Main processing step"""
        debug.trace_fmtd(5, "Script.run_main_step(): self={s}", s=self)

        # Validate input files
        self.check_regular_file(self.main_filename)
        if system.is_directory(self.other_filename):
            self.other_filename = gh.form_path(self.other_filename,
                                               gh.basename(self.main_filename))
        self.check_regular_file(self.other_filename)

        # Find the backup baseline if not specified
        if not self.backup_filename:
            EPSILON = 1e-6
            main_timestamp = get_numeric_timestamp(self.main_filename)
            other_timestamp = get_numeric_timestamp(self.other_filename)
            min_timestamp = min(main_timestamp, other_timestamp) - EPSILON

            # Find most recent backup older than the two files
            # TODO: reinstate ~ check (or ./backup/f ./backup/~ and backup/f.~N~)
            SEP = "\t"
            backup_files = gh.get_files_matching_specs("{d}/{b}{SEP}{d}/{b}*".
                                                       format(d=self.backup_dir,
                                                              b=self.basename, SEP=SEP)
                                                       .split(SEP))
            timestamped_backups = [(f, ts) for (f, ts) in sorted(zip(backup_files,
                                                                     map(get_numeric_timestamp, backup_files)),
                                                                 key=lambda f_ts: f_ts[1], reverse=True)
                                   if ((ts <= min_timestamp) or IGNORE_TIMESTAMP)]
            debug.trace_expr(5, timestamped_backups)
            if timestamped_backups:
                self.backup_filename = timestamped_backups[0][0]

            # If no suitable backup, use the older of the two files
            if ((not self.backup_filename) and self.skip_baseline):
                self.backup_filename = self.main_filename if (main_timestamp < other_timestamp) else self.other_filename
        if not self.backup_filename:
            system.exit("Error: Unable to find baseline backup for {f1} ({ts1}) and {f2} ({ts2}) ",
                        f1=self.main_filename, ts1=get_timestamp(self.main_filename),
                        f2=self.other_filename, ts2=get_timestamp(self.other_filename))
        self.check_regular_file(self.backup_filename)

        # Do the merge and check for conflicts
        # NOTE: Makes copy of each before merge in case inadvertantly clobbered (e.g., $TMP/filename1.copy)
        # TODO: produce report including diff listings of old vs. new
        system.create_directory(self.temp_base)
        temp_new_file = gh.form_path(self.temp_base, self.basename)
        gh.copy_file(self.main_filename, gh.form_path(self.temp_base, gh.basename(self.main_filename) + ".copy"))
        gh.copy_file(self.other_filename, gh.form_path(self.temp_base, gh.basename(self.other_filename) + ".copy"))
        result = gh.run("{m} {f1} {b} {f2} > {new}", m=MERGE, f1=self.main_filename,
                        b=self.backup_filename, f2=self.other_filename, new=temp_new_file)
        if re.search(r"conflict in merge", result):
            system.print_stderr("Error in merge:\n\t{r}", r=gh.indent_lines(result))
        elif self.update_file1:
            if not self.quiet:
                print("Updating {f1}, after rename to {f1}.backup".format(f1=self.main_filename))
            gh.rename_file(self.main_filename, (self.main_filename + ".backup"))
            gh.copy_file(temp_new_file, self.main_filename)
        elif self.use_stdout:
            print(system.read_entire_file(temp_new_file))
        else:
            if not self.quiet:
                print(f"Merge result:\n\t{result}")
                debug.trace(4, f"See {temp_new_file}")
        return

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=debug.QUITE_DETAILED)
    # TODO: add examples (especially with consolidated backup directory)
    app = Script(
        description=__doc__.format(prog=gh.basename(__file__)),
        skip_input=True,
        manual_input=True,
        use_temp_base_dir=True,
        boolean_options=[(IGNORE_ERRORS, "Ignore errors in processing"),
                         (UPDATE_MAIN_FILE, f"Replace {MAIN_FILENAME} with merged result--be careful"),
                         (USE_STDOUT, "Use standard output for result"),
                         (QUIET, "Omit status messages")],
        positional_arguments=[MAIN_FILENAME, OTHER_FILENAME],
        text_options=[(BACKUP_FILENAME, "Backup file for baseline"),
                      (BACKUP_DIR, "Backup directory"),
                      (BASENAME, f"Basename for backup check--if not {MAIN_FILENAME}")])
    app.run()



#! /usr/bin/env python3
# 
# Sample script using Main class. By default, this outputs lines that contain
# "fubar". There is an option to check for lines matching a regular expression.
#

"""Simple illustration of Main class

Sample usage:
   echo $'foobar\\nfubar' | {script} --check-fubar -
"""

# Standard packages
## TODO: from collections import defaultdict

# Local packages
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla.my_regex import my_re

# Constants
TL = debug.TL
FUBAR_ARG = "check-fubar"
REGEX_ARG = "regex"
ENTIRE_ARG = "entire-file"
PARA_ARG = "para"

class SimpleFilter:
    """Regex filter for text"""

    def __init__(self, regex=None, flags=0):
        """Initialize class with PATTERN which can be a REGEX"""
        self.regex = regex
        self.flags = flags
        debug.trace_object(TL.DETAILED, self, label=f"{self.__class__.__name__} instance")

    def include(self, line):
        """Output line if filter met"""
        include = my_re.search(self.regex, line, flags=self.flags)
        debug.trace(TL.QUITE_VERBOSE, "SimpleFilter.include({line!r}) => {include}")
        return include

    
class Script(Main):
    """Input processing class"""
    filter_inst = None

    def setup(self):
        """Check results of command line processing"""
        debug.trace_fmtd(TL.DETAILED, "Script.setup(): self={s}", s=self)
        regex = self.get_parsed_option(REGEX_ARG, "")
        check_fubar = self.get_parsed_option(FUBAR_ARG, not regex)
        slurp_mode = self.get_parsed_option(ENTIRE_ARG)
        para_mode = self.get_parsed_option(PARA_ARG)
        debug.assertion(bool(regex) ^ bool(check_fubar))
        if check_fubar:
            regex = "fubar"
        if slurp_mode:
            self.file_input_mode = True
        if para_mode:
            self.paragraph_mode = True
        self.filter_inst = SimpleFilter(regex)
        debug.trace_object(TL.VERBOSE, self, label=f"{self.__class__.__name__} instance")

    def process_line(self, line):
        """Processes current LINE from input
        Note: can be a paragraph or an entire file (a la perl)"""
        debug.trace(TL.QUITE_DETAILED, f"Script.process_line({line!r})")
        if self.filter_inst.include(line):
            print(line)
        return

def main():
    """Entry point"""
    app = Script(
        description=__doc__.format(script=gh.basename(__file__)),
        boolean_options=[(FUBAR_ARG, "Check for 'fubar' in line"),
                         (PARA_ARG, "Process file in paragraphs (a la Perl)"),
                         (ENTIRE_ARG, "Apply filter to entire file (a la perl slurping")],
        text_options=[(REGEX_ARG, "Regular expression to check")])
    app.run()

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_VERBOSE)
    main()



#! /usr/bin/env python3
#
# Test(s) for ../ipython_utils.py
#
# Notes:
# - For debugging the tested script, the ALLOW_SUBCOMMAND_TRACING environment
#   option shows tracing output normally suppressed by unittest_wrapper.py.
# - This can be run as follows:
#   $ PYTHONPATH=".:$PYTHONPATH" python ./mezcla/tests/test_ipython_utils.py
#

"""Tests for ipython_utils module"""

# Standard packages
## TODO: from collections import defaultdict

# Installed packages
import pytest

# Local packages
from mezcla.unittest_wrapper import TestWrapper, invoke_tests
from mezcla import debug
import mezcla.ipython_utils as THE_MODULE

#------------------------------------------------------------------------

class TestIt(TestWrapper):
    """Class for command-line based testcase definition"""
    script_module = TestWrapper.get_testing_module_name(__file__, THE_MODULE)

    def test_grep_obj_methods(self):
        """Test (module) object grepping"""
        debug.trace(4, f"TestIt.test_grep_obj_methods(); self={self}")
        import_functions = THE_MODULE.grep_obj_methods(THE_MODULE, "import_*")
        assert ("import_module_globals" in import_functions)
        leading_space_names = THE_MODULE.grep_obj_methods(THE_MODULE, "^ +")
        assert (not leading_space_names)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_import_module_globals(self):
        """Test module globals import"""
        debug.trace(4, f"TestIt.test_import_module_globals(); self={self}")
        #
        # Import this module (i.e., "mezcla.ipython_utils")
        globals_dict = {}
        THE_MODULE.import_module_globals(self.script_module, globals_dict=globals_dict)
        assert ("TL" in globals_dict)
        assert ("import_module_globals" in globals_dict)
        #
        # note: following fails because the module argument needs to be text (as above)
        globals_dict = {}
        THE_MODULE.import_module_globals(THE_MODULE, globals_dict=globals_dict, ignore_errors=True)
        assert (not globals_dict)
        return

    @pytest.mark.xfail                   # TODO: remove xfail
    def test_pr_dir(self):
        """Makes sure object directory printed"""
        # Sample output:
        #   In [67]: pr_dir(Dummy())
        #   ['__class__', '__delattr__', '__dict__', '__dir__', '__doc__', '__eq__', '__format__', '__ge__', '__getattribute__', '__getstate__', '__gt__', '__hash__', '__init__', '__init_subclass__', '__le__', '__lt__', '__module__', '__ne__', '__new__', '__reduce__', '__reduce_ex__', '__repr__', '__setattr__', '__sizeof__', '__str__', '__subclasshook__', '__weakref__', 'dummy_member']
        #
        class Dummy():
            """Dummy class for testing pr_dir"""
            dummy_member = True
        #
        dummy_instance = Dummy()
        THE_MODULE.pr_dir(dummy_instance)
        stdout = self.get_stdout()
        assert "dummy_member" in stdout
        assert "'__class__'" in stdout

#------------------------------------------------------------------------

if __name__ == '__main__':
    debug.trace_current_context()
    invoke_tests(__file__)



#! /usr/bin/env python3
#
# Illustrates how to use Stable Diffusion via Hugging Face (HF) diffusers package,
# including gradio-based UI for txt2img and img2img. This also uses the clip_interrogator
# package for img2txt.
#
# Details:
# - via https://huggingface.co/spaces/stabilityai/stable-diffusion
#   also uses https://huggingface.co/CompVis/stable-diffusion-v1-4
#   and https://stackoverflow.com/questions/48273205/accessing-incoming-post-data-in-flask
#
# - This was designed originally for text-to-image (i.e., from prompt to image).
#   However, it has been adapted to support image-to-image as well, which includes an image
#   input along the prompt(s).
#
# - Support is also included for clip interrogation for image-to-text, which is not yet
#   part of the HF diffusers API.
#
# Note:
# - For tips on parameter settings, see
#   https://getimg.ai/guides/interactive-guide-to-stable-diffusion-guidance-scale-parameter
# - For a full-featured interface to Stable Diffusion, see
#   https://github.com/AUTOMATIC1111/stable-diffusion-webui
# - For a stylish Stable Diffusion interface, see
#   https://github.com/comfyanonymous/ComfyUI
#
# TODO2:
# - See why HF keeps downloading model (e.g., make sure cache permanent).
# - Add test cases for flask-based API.
# TODO:
# - Set GRADIO_SERVER_NAME to 0.0.0.0?
#

"""Image generation via HF Stable Diffusion (SD) API"""

# Standard modules
import base64
from io import BytesIO
import json
import time

# Installed modules
import diskcache
from flask import Flask, request
## TODO: see why following needed (i.e., plain PIL import yields intermittent errors)
import PIL.Image
import requests
gr = None

# Local modules
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla.my_regex import my_re
from mezcla import system

# Constants/globals
TL = debug.TL
PROMPT = system.getenv_text("PROMPT", "your favorite politician in a tutu",
                            "Textual prompt describing image")
NEGATIVE_PROMPT = system.getenv_text("NEGATIVE_PROMPT", "photo realistic",
                            "Negative tips for image")
GUIDANCE_HELP = "Degree of fidelity to prompt (1-to-30 w/ 7 suggested)--higher for more; aka Classifier Free Guidance (CFG)"
GUIDANCE_SCALE = system.getenv_int("GUIDANCE_SCALE", 7,
                                   description=GUIDANCE_HELP)
SD_URL_ENV = system.getenv_value("SD_URL", None,
                             "URL for SD TCP/restful server--new via flask or remote")
SD_URL = (SD_URL_ENV if (SD_URL_ENV is not None) and (SD_URL_ENV.strip() not in ["", "-"]) else None)
SD_PORT = system.getenv_int("SD_PORT", 9700,
                            "TCP port for SD server")
SD_DEBUG = system.getenv_int("SD_DEBUG", False,
                             "Run SD server in debug mode")
USE_HF_API = system.getenv_bool("USE_HF_API", not SD_URL,
                                "Use Huggingface API instead of TCP server")
CHECK_UNSAFE = system.getenv_bool("CHECK_UNSAFE", False,
                                  "Apply unsafe word list regex filter")
NUM_IMAGES = system.getenv_int("NUM_IMAGES", 1,
                               "Number of images to generated")
BASENAME = system.getenv_text("BASENAME", "sd-app-image",
                              "Basename for saving images")
FULL_PRECISION = system.getenv_bool("FULL_PRECISION", False,
                                    "Use full precision GPU computations")
LOW_MEMORY = system.getenv_bool("LOW_MEMORY", (not FULL_PRECISION),
                                "Use low memory computations such as via float16")
DUMMY_RESULT = system.getenv_bool("DUMMY_RESULT", False,
                                  "Mock up SD server result")
DISK_CACHE = system.getenv_value("SD_DISK_CACHE", None,
                                 "Path to directory with disk cache")
USE_IMG2IMG = system.getenv_bool("USE_IMG2IMG", False,
                                 "Use image-to-image instead of text-to-image")
USE_IMG2TXT = system.getenv_bool("USE_IMG2TXT", False,
                                 "Use image-to-text instead of image generation")
DENOISING_FACTOR = system.getenv_float("DENOISING_FACTOR", 0.75,
                                       "How much of the input image to randomize--higher for more")
HF_SD_MODEL = system.getenv_text(
    "HF_SD_MODEL", "CompVis/stable-diffusion-v1-4",
    description="Hugging Face model for Stable Diffusion")
STREAMLINED_CLIP = system.getenv_bool(
    # note: Doesn't default to LOW_MEMORY, which just uses 16-bit floating point:
    # several settings are changed (see apply_low_vram_defaults in clip_interrogator).
    "STREAMLINED_CLIP", False,
    description="Use streamlined CLIP settings to reduce memory usage")
REGULAR_CLIP = (not STREAMLINED_CLIP)
CAPTION_MODEL = system.getenv_text(
    "CAPTION_MODEL", ("blip-large" if REGULAR_CLIP else "blip-base"),
    description="Caption model to use in CLIP interrogation")
CLIP_MODEL = system.getenv_text(
    "CLIP_MODEL", ("ViT-L-14/openai" if REGULAR_CLIP else "ViT-B-16/openai"),
    # TODO4: see https://arxiv.org/pdf/2010.11929.pdf for ViT-S-NN explanation
    description="Model to use for CLIP interrogation")
#
BATCH_ARG = "batch"
SERVER_ARG = "server"
UI_ARG = "UI"
PORT_ARG = "port"
PROMPT_ARG = "prompt"
NEGATIVE_ARG = "negative"
GUIDANCE_ARG = "guidance"
TXT2IMG_ARG = "txt2img"
IMG2IMG_ARG = "img2img"
IMG2TXT_ARG = "img2txt"
DENOISING_ARG = "denoising-factor"
DUMMY_BASE64_IMAGE = "iVBORw0KGgoAAAANSUhEUgAAAA8AAAAPAgMAAABGuH3ZAAAADFBMVEUAAMzMzP////8AAABGA1scAAAAJUlEQVR4nGNgAAFGQUEowRoa6sCABBZowAgsgBEIGUQCRALAPACMHAOQvR4HGwAAAABJRU5ErkJggg=="
DUMMY_IMAGE_FILE = gh.resolve_path("dummy-image.png")
HTTP_OK = 200
# note: RestFUL API keys (via flask)
IMAGES_KEY = "images"
CAPTION_KEY = "caption"

#--------------------------------------------------------------------------------
# Globals
# note: includes support for conditional imports for HF/PyTorch

torch = None
load_dataset = None

word_list = []

sd_instance = None
flask_app = Flask(__name__)


#--------------------------------------------------------------------------------
# Utility function (TODO2: put with others)

def init():
    """Initialize Hugging face for Stable Diffusion"""
    # TODO2: merge with init_stable_diffusion
    debug.trace(4, "in hf_stable_diffusion.init")

    # Load Hugging Face and PyTorch
    global load_dataset, torch
    if USE_HF_API:
        # pylint: disable=import-outside-toplevel, import-error, redefined-outer-name
        from datasets import load_dataset
        import torch

    # Load blacklist for prompt terms
    global word_list
    if CHECK_UNSAFE:
        word_list_dataset = load_dataset("stabilityai/word-list", data_files="list.txt", use_auth_token=True)
        word_list = word_list_dataset["train"]['text']
        debug.trace_expr(5, word_list)
    debug.trace(5, "out hf_stable_diffusion.init")


def show_gpu_usage(level=TL.DETAILED):
    """Show usage for GPU memory, etc.
    TODO: support other types besides NVidia"""
    if USE_HF_API:
        debug.trace(level, "GPU usage")
        debug.trace(level, gh.run("nvidia-smi"))

#-------------------------------------------------------------------------------
# Main Stable Diffusion support

class StableDiffusion:
    """Class providing Stable Diffusion generative AI (e.g., text-to-image)"""

    def __init__(self, use_hf_api=None, server_url=None, server_port=None, low_memory=None):
        ## BAD: debug.trace(4, f"{self.__class__.__name__}.__init__{(use_hf_api, server_url, server_port)}")
        ## TODO2: derive class name from call stack
        class_name = "StableDiffusion"
        debug.trace(4, f"{class_name}.__init__{(use_hf_api, server_url, server_port)}")
        if use_hf_api is None:
            use_hf_api = USE_HF_API
        self.use_hf_api = use_hf_api
        if (server_url is None) and (not self.use_hf_api):
            server_url = SD_URL
        self.server_url = server_url
        if server_port is None:
            server_port = SD_PORT
        if self.server_url and not my_re.search(r"^https?", self.server_url):
            # note: remote flask server (e.g., on GPU server)
            self.server_url = f"http://{self.server_url}"
            debug.trace(4, f"Added http protocol to URL: {self.server_url}")
        if self.server_url:
            if not my_re.search(r":\d+", self.server_url):
                # TODO3: http://base-url/path => http://base-url:port/path
                self.server_url += f":{server_port}"
            elif not server_port:
                system.print_stderr(f"Warning: ignoring port {server_port} as already in URL {self.server_url}")
        elif self.use_hf_api:
            pass
        else:
            debug.trace(4,"Warning: Unexpected condition in {class_name}.__init__: no server_url")
        if low_memory is None:
            low_memory = LOW_MEMORY
        self.low_memory = low_memory
        self.pipe = None
        self.img2txt_engine = None
        self.cache = None
        if DISK_CACHE:
            self.cache = diskcache.Cache(
                DISK_CACHE,                   # path to dir
                disk=diskcache.core.JSONDisk, # avoid serialization issue
                disk_compress_level=0,        # no compression
                cull_limit=0)                 # no automatic pruning
        debug.assertion(bool(self.use_hf_api) != bool(self.server_url))
        show_gpu_usage()
        debug.trace_object(5, self, label=f"{class_name} instance")
    
    def init_pipeline(self, txt2img=None, img2img=None):
        """Initialize Stable Diffusion"""
        debug.trace(4, "init_pipeline()")
        debug.assertion(not (txt2img and img2img))
        ## TODO2: fix lack of support for img2img
        debug.assertion(not img2img)
        # pylint: disable=import-outside-toplevel
        from diffusers import StableDiffusionPipeline
        model_id = HF_SD_MODEL
        device = "cuda"
        # TODO2: automatically use LOW_MEMORY if GPU memory below 8gb
        dtype=(torch.float16 if self.low_memory else None)
        pipe = StableDiffusionPipeline.from_pretrained(model_id, torch_dtype=dtype)
        debug.trace_expr(5, pipe, dtype)
        pipe = pipe.to(device)
        pipe.set_progress_bar_config(disable=True)
        pipe.enable_attention_slicing()
        debug.trace_object(6, pipe)
        show_gpu_usage()
        return pipe

    def init_txt2img(self):
        """Initialize Stable Diffusion text-to-image support (i.e., txt2img)"""
        debug.trace(4, "init_txt2img()")
        return self.init_pipeline(txt2img=True)

    def init_img2img(self):
        """Initialize Stable Diffusion image-to-image support (i.e., img2img)"""
        debug.trace(4, "init_img2img()")
        # pylint: disable=import-outside-toplevel
        from diffusers import StableDiffusionImg2ImgPipeline
        # TODO2: v1-5
        model_id = HF_SD_MODEL
        device = "cuda"
        # TODO2: automatically use LOW_MEMORY if GPU memory below 8gb
        dtype=(torch.float16 if self.low_memory else None)
        pipe = StableDiffusionImg2ImgPipeline.from_pretrained(model_id, torch_dtype=dtype)
        debug.trace_expr(5, pipe, dtype)
        pipe = pipe.to(device)
        pipe.set_progress_bar_config(disable=True)
        pipe.enable_attention_slicing()
        debug.trace_object(6, pipe)
        show_gpu_usage()
        return pipe

    def init_clip_interrogation(self):
        """Initialize CLIP interrogation for use with Stable Diffusion"""
        # pylint: disable=import-outside-toplevel
        debug.trace(4, "init_clip_interrogation()")
        from clip_interrogator import Config, Interrogator
        clip_config = Config(caption_model_name=CAPTION_MODEL,
                             clip_model_name=CLIP_MODEL)
        if STREAMLINED_CLIP:
            clip_config.apply_low_vram_defaults()
        self.img2txt_engine = Interrogator(clip_config)

    def infer(self, prompt=None, negative_prompt=None, scale=None, num_images=None,
              skip_img_spec=False, width=None, height=None, skip_cache=False):
        """Generate images using positive PROMPT and NEGATIVE one, along with guidance SCALE,
        and targetting a WIDTHxHEIGHT image.
        Returns list of NUM image specifications in base64 format (e.g., for use in HTML).
        Note: If SKIP_IMG_SPEC specified, result is formatted for HTML IMG tag.
        If SKIP_CACHE, then new results are always generated.
        """
        debug.trace_expr(4, prompt, negative_prompt, scale, num_images, skip_img_spec, skip_cache, prefix=f"\nin {self.__class__.__name__}.infer:\n\t", delim="\n\t",  suffix="}\n", max_len=1024)
        if num_images is None:
            num_images = NUM_IMAGES
        if scale is None:
            scale = GUIDANCE_SCALE
        for prompt_filter in word_list:
            if my_re.search(rf"\b{prompt_filter}\b", prompt):
                raise RuntimeError("Unsafe content found. Please try again with different prompts.")
    
        images = []
        params = (prompt, negative_prompt, scale, num_images, skip_img_spec, width, height)

        if ((self.cache is not None) and (not skip_cache)):
            images = self.cache.get(params)
        if images and len(images) > 0:
            ## TEST:
            ## debug.trace_fmt(6, "Using cached result for params {p}: ({r})",
            ##                 p=params, r=images)
            debug.trace_fmt(5, "Using cached infer result: ({r})", r=images)
        else:
            images = self.infer_non_cached(*params)
            if self.cache is not None:
                self.cache.set(params, images)
                debug.trace_fmt(6, "Setting cached result (r={r})", r=images)
        return images

    def infer_non_cached(self, prompt=None, negative_prompt=None, scale=None, num_images=None,
                         skip_img_spec=False, width=None, height=None):
        """Non-cached version of infer"""
        params = (prompt, negative_prompt, scale, num_images, skip_img_spec, width, height)
        debug.trace(5, f"{self.__class__.__name__}.infer_non_cached{params}")
        images = []
        if self.use_hf_api:
            if DUMMY_RESULT:
                result = [DUMMY_BASE64_IMAGE]
                debug.trace(5, f"early exit infer_non_cached() => {result}")
                return result

            if not self.pipe:
                self.pipe = self.init_pipeline()
            start_time = time.time()
            image_info = self.pipe(prompt, negative_prompt=negative_prompt, guidance_scale=scale,
                                   num_images_per_prompt=num_images, width=width, height=height)
            debug.trace_expr(4, image_info)
            debug.trace_object(5, image_info, "image_info")
            num_generated = 0
            for _i, image in enumerate(image_info.images):
                debug.trace_expr(4, image)
                debug.trace_object(5, image, "image")
                b64_encoding = image
                debug.assertion(isinstance(image, PIL.Image.Image))
                try:
                    num_generated += 1
                    b64_encoding = encode_PIL_image(image)
                    if not skip_img_spec:
                        b64_encoding = (f"data:image/png;base64,{b64_encoding}")
                except:
                    system.print_exception_info("image-to-base64")
                images.append(b64_encoding)
            elapsed = round(time.time() - start_time, 3)
            debug.trace(4, f"{elapsed} seconds to generate {num_images} images")
            debug.assertion(num_generated == num_images)
            show_gpu_usage()
        else:
            debug.assertion(self.server_url)
            url = self.server_url
            payload = {'prompt': prompt, 'negative_prompt': negative_prompt, 'scale': scale,
                       'num_images': num_images}
            images_request = requests.post(url, json=payload, timeout=(5 * 60))
            debug.trace_object(6, images_request)
            debug.trace_expr(5, payload, images_request, images_request.json(), delim="\n")
            for image in images_request.json()[IMAGES_KEY]:
                image_b64 = image
                if not skip_img_spec:
                    image_b64 = (f"data:image/png;base64,{image_b64}")
                images.append(image_b64)
        result = images
        debug.trace_fmt(5, "infer_non_cached() => {r!r}", r=result)
        return result

    def infer_img2img(self, image_b64=None, denoise=None,  prompt=None, negative_prompt=None, scale=None, num_images=None,
                      skip_img_spec=False, skip_cache=False):
        """Generate images from IMAGE_B64 using positive PROMPT and NEGATIVE one, along with guidance SCALE and NUM_IMAGES
        Returns list of NUM image specifications in base64 format (e.g., for use in HTML).
        Note: If SKIP_IMG_SPEC specified, result is formatted for HTML IMG tag.
        If SKIP_CACHE, then new results are always generated.
        """
        debug.trace_expr(4, image_b64, denoise, prompt, negative_prompt, scale, num_images, skip_img_spec, skip_cache, prefix=f"\nin {self.__class__.__name__}.infer_img2img: {{\n\t", delim="\n\t", suffix="}\n", max_len=1024)
        if num_images is None:
            num_images = NUM_IMAGES
        if scale is None:
            scale = GUIDANCE_SCALE
        if denoise is None:
            denoise = DENOISING_FACTOR
        for prompt_filter in word_list:
            if my_re.search(rf"\b{prompt_filter}\b", prompt):
                raise RuntimeError("Unsafe content found. Please try again with different prompts.")
    
        images = []
        params = (image_b64, denoise, prompt, negative_prompt, scale, num_images, skip_img_spec)

        if ((self.cache is not None) and (not skip_cache)):
            images = self.cache.get(params)
        if images and len(images) > 0:
            ## TEST:
            ## debug.trace_fmt(6, "Using cached result for params {p}: ({r})",
            ##                 p=params, r=images)
            debug.trace_fmt(5, "Using cached infer result: ({r})", r=images)
        else:
            images = self.infer_img2img_non_cached(*params)
            if self.cache is not None:
                self.cache.set(params, images)
                debug.trace_fmt(6, "Setting cached result (r={r})", r=images)
        return images

    def infer_img2img_non_cached(self, image_b64=None, denoise=None, prompt=None, negative_prompt=None, scale=None, num_images=None,
                                 skip_img_spec=False):
        """Non-cached version of infer_img2img"""
        params = (image_b64, denoise, prompt, negative_prompt, scale, num_images, skip_img_spec)
        params_spec = params
        if debug.detailed_debugging():
            params_spec = tuple(map(gh.elide, params))
        debug.trace(5, f"{self.__class__.__name__}.infer_img2img_non_cached{params_spec}")
        images = []
        if self.use_hf_api:
            if DUMMY_RESULT:
                result = ["iVBORw0KGgoAAAANSUhEUgAAAA8AAAAPAgMAAABGuH3ZAAAADFBMVEUAAMzMzP////8AAABGA1scAAAAJUlEQVR4nGNgAAFGQUEowRoa6sCABBZowAgsgBEIGUQCRALAPACMHAOQvR4HGwAAAABJRU5ErkJggg=="]
                debug.trace(5, f"early exit infer_img2img_non_cached() => {result}")
                return result

            if not self.pipe:
                self.pipe = self.init_img2img()
            # Get input image
            input_image = create_image(decode_base64_image(image_b64))
            # Generate derived output images(s)
            image_info = self.pipe(image=input_image, strength=denoise, prompt=prompt, negative_prompt=negative_prompt, guidance_scale=scale,
                                   num_images_per_prompt=num_images)
            debug.trace_expr(4, image_info)
            debug.trace_object(5, image_info, "image_info")
            start_time = time.time()
            num_generated = 0
            for _i, image in enumerate(image_info.images):
                debug.trace_expr(4, image)
                debug.trace_object(5, image, "image")
                b64_encoding = image
                debug.assertion(isinstance(image, PIL.Image.Image))
                try:
                    num_generated += 1
                    b64_encoding = encode_PIL_image(image)
                    if not skip_img_spec:
                        b64_encoding = (f"data:image/png;base64,{b64_encoding}")
                except:
                    system.print_exception_info("image-to-base64")
                images.append(b64_encoding)
            elapsed = round(time.time() - start_time, 3)
            debug.trace(4, f"{elapsed} seconds to derive {num_images} images")
            debug.assertion(num_generated == num_images)
            show_gpu_usage()
        else:
            debug.assertion(self.server_url)
            url = self.server_url
            payload = {'image_b64': image_b64, 'strength': denoise, 'prompt': prompt, 'negative_prompt': negative_prompt, 'scale': scale,
                       'num_images': num_images}
            images_request = requests.post(url, json=payload, timeout=(5 * 60))
            debug.trace_object(6, images_request)
            debug.trace_expr(5, payload, images_request, images_request.json(), delim="\n")
            for image in images_request.json()[IMAGES_KEY]:
                image_b64 = image
                if not skip_img_spec:
                    image_b64 = (f"data:image/png;base64,{image_b64}")
                images.append(image_b64)
        result = images
        debug.trace_fmt(5, "infer_img2img_non_cached() => {r!r}", r=result)
        
        return result

    def infer_img2txt(self, image_b64=None, skip_cache=False):
        """Return likely caption text for IMAGE_B64 in base64 encoding
        Note: If SKIP_CACHE, then new results are always generated.
        """
        debug.trace_fmt(4, f"infer_img2txt({gh.elide(image_b64)}, sk={skip_cache})")
        params = (image_b64)
        description = None
        if ((self.cache is not None) and (not skip_cache)):
            description = self.cache.get(params)
        if (description is not None):
            debug.trace_fmt(5, "Using cached infer_img2txt result: ({r!r})", r=description)
        else:
            description = self.infer_img2txt_non_cached(image_b64)
            if self.cache is not None:
                self.cache.set(params, description)
                debug.trace_fmt(6, "Setting cached result (r={r!r})", r=description)
        # Make sure a string
        if (description is None):
            description = ""

        return description
    
    def infer_img2txt_non_cached(self, image_b64=None):
        """Non-cached version of infer_img2txt"""
        image_caption = ""
        if (self.use_hf_api):
            # Get input image and infer likely caption text
            image = create_image(decode_base64_image(image_b64))
            if not self.img2txt_engine:
                self.init_clip_interrogation()
            image_caption = self.img2txt_engine.interrogate(image)
        else:
            debug.assertion(self.server_url)
            url = self.server_url
            payload = {'image_b64': image_b64}
            request_result = requests.post(url, json=payload, timeout=(5 * 60))
            debug.trace_object(6, request_result)
            debug.trace_expr(5, payload, request_result, request_result.json(), delim="\n")
            image_caption = request_result.json()[CAPTION_KEY]
            
        debug.trace_fmt(5, "infer_img2txt_non_cached() => {r!r}", r=image_caption)
        return image_caption
    
#-------------------------------------------------------------------------------
# Middleware

def init_stable_diffusion(use_hf_api=None):
    """Initialize stable diffusion usage, locally if USE_HF_API"""
    debug.trace(4, f"init_stable_diffusion({use_hf_api})")
    init()
    global sd_instance
    sd_instance = StableDiffusion(use_hf_api=use_hf_api)
    debug.trace_expr(5, sd_instance)


@flask_app.route('/', methods=['GET', 'POST'])
def handle_infer():
    """Process request to do inference to generate image from text"""
    # Note: result return via hash with images key
    debug.trace(6, "[flask_app /] handle_infer()")
    # TODO3: request => flask_request
    debug.trace_object(5, request)
    params = request.get_json()
    debug.trace_expr(5, params)
    if not sd_instance:
        init_stable_diffusion()
    images_spec = {IMAGES_KEY: sd_instance.infer(**params)}
    # note: see https://stackoverflow.com/questions/45412228/sending-json-and-status-code-with-a-flask-response
    result = (json.dumps(images_spec), HTTP_OK)
    debug.trace_object(7, result)
    debug.trace_fmt(7, "handle_infer() => {r}", r=result)
    return result


@flask_app.route('/txt2img', methods=['GET', 'POST'])
def handle_infer_txt2img():
    """[Alias] Process request to do inference to generate image from text"""
    debug.trace(6, "[flask_app /] handle_infer_txt2img()")
    return handle_infer()
    

def infer(prompt=None, negative_prompt=None, scale=None, num_images=None, skip_img_spec=None):
    """Wrapper around StableDiffusion.infer()
    Note: intended just for the gradio UI"
    """
    debug.trace(5, f"[sd_instance] infer{(prompt, negative_prompt, scale, skip_img_spec)}")
    if not sd_instance:
        init_stable_diffusion()
    return sd_instance.infer(prompt=prompt, negative_prompt=negative_prompt, scale=scale, num_images=num_images, skip_img_spec=skip_img_spec)


@flask_app.route('/img2img', methods=['GET', 'POST'])
def handle_infer_img2img():
    """Process request to do inference to generate similar image from existing image"""
    # Note: result return via hash with images key
    debug.trace(6, "[flask_app /] handle_infer_img2img()")
    # TODO3: request => flask_request
    debug.trace_object(5, request)
    params = request.get_json()
    debug.trace_expr(5, params)
    if not sd_instance:
        init_stable_diffusion()
    images_spec = {IMAGES_KEY: sd_instance.infer_img2img(**params)}
    # note: see https://stackoverflow.com/questions/45412228/sending-json-and-status-code-with-a-flask-response
    result = (json.dumps(images_spec), HTTP_OK)
    debug.trace_object(7, result)
    debug.trace_fmt(7, "handle_infer_img2img() => {r}", r=result)
    return result


def infer_img2img(image_spec=None, denoise=None,  prompt=None, negative_prompt=None, scale=None, num_images=None, skip_img_spec=None):
    """Wrapper around StableDiffusion.infer_img2img()
    Note: intended just for the gradio UI"
    """
    debug.trace(5, f"[sd_instance] infer_img2img{(gh.elide(image_spec), denoise, prompt, negative_prompt, scale, skip_img_spec)}")
    if isinstance(image_spec, list):
        debug.trace(5, "Warning: using first image in image_spec for infer_img2img")
        image_spec = image_spec[0]
    if ((image_spec is not None) and (not isinstance(image_spec, str))):
        debug.trace_expr(7, image_spec)
        image = (image_spec)
        image_spec = encode_PIL_image(image)
    image_b64 = image_spec        
    if not sd_instance:
        init_stable_diffusion()
    return sd_instance.infer_img2img(image_b64=image_b64, denoise=denoise, prompt=prompt, negative_prompt=negative_prompt, scale=scale, num_images=num_images, skip_img_spec=skip_img_spec)


@flask_app.route('/img2txt', methods=['GET', 'POST'])
def handle_infer_img2txt():
    """Process request to do inference to generate text description of image
    Note: result returned via hash with caption key"""
    ## TODO3: add helper for common flask bookkeeping
    debug.trace(6, "[flask_app /] handle_infer_img2txt()")
    debug.trace_object(5, request)
    params = request.get_json()
    debug.trace_expr(5, params)
    if not sd_instance:
        init_stable_diffusion()
    caption_spec = {CAPTION_KEY: sd_instance.infer_img2txt(**params)}
    result = (json.dumps(caption_spec), HTTP_OK)    
    debug.trace_object(7, result)
    debug.trace_fmt(7, "handle_infer_img2txt() => {r}", r=result)
    return result


def infer_img2txt(image_spec):
    """Wrapper around StableDiffusion.infer_img2txt()
    Note: intended just for the gradio UI"
    """
    debug.trace(5, f"[sd_instance] infer_img2txt({gh.elide(image_spec)})")
    image_b64 = image_spec
    ## TEMP:
    if isinstance(image_b64, PIL.Image.Image):
        image_b64 = encode_PIL_image(image_b64)
    if not sd_instance:
        init_stable_diffusion()
    return sd_instance.infer_img2txt(image_b64)

#--------------------------------------------------------------------------------
# Utility functions

def encode_image_data(image_bytes):
    """Convert IMAGE_BYTES to base64 string"""
    debug.assertion(isinstance(image_bytes, bytes))
    result = base64.b64encode(image_bytes).decode()
    debug.trace(6, f"encode_image_data({gh.elide(image_bytes)}) => {gh.elide(result)}")
    return result


def encode_image_file(filename):
    """Encode image in FILENAME via base64 string"""
    binary_data = system.read_binary_file(filename)
    result = encode_image_data(binary_data)
    debug.trace(6, f"encode_image_file({filename}) => {gh.elide(result)}")
    return result


def encode_PIL_image(image):
    """Convert from PIL image into base64"""
    debug.assertion(isinstance(image, PIL.Image.Image))
    ## BAD: result = encode_image_data(image.tobytes())
    ## note: all sorts of silly issues with PIL!
    bytes_fh = BytesIO()
    image.save(bytes_fh, format="PNG")
    bytes_fh.seek(0)
    result = encode_image_data(bytes_fh.read())
    debug.trace(6, f"encode_PIL_image({gh.elide(image)}) => {gh.elide(result)}")
    return result


def decode_base64_image(image_encoding):
    """Decode IMAGE_ENCODING from base64 returning bytes"""
    # note: "encodes" UTF-8 text of base-64 encoding as bytes object for str, and then decodes into image bytes
    result = base64.decodebytes(image_encoding.encode())
    debug.trace(6, f"decode_base64_image({gh.elide(image_encoding)}) => {gh.elide(result)}")
    return result

def create_image(image_data):           # TODO1: create_PIL_image
    """Create PIL image from IMAGE_DATA bytes"""
    ## TODO4?: create_PIL_image
    result = PIL.Image.open(BytesIO(image_data)).convert("RGB")
    debug.trace(6, f"create_image({gh.elide(image_data)!r}) => {result}")
    return result

def write_image_file(filename, image_spec):
    """Write to FILENAME the base64 data in IMAGE_SPEC"""
    debug.trace(5, f"write_image_file({filename}, {gh.elide(image_spec)})")
    system.write_binary_file(filename, decode_base64_image(image_spec))

#-------------------------------------------------------------------------------
# User interface

def upload_image(upload_control):
    """Upload image data from UPLOAD_CONTROL returning base64 encoded image"""
    debug.trace(4, f"upload_image({upload_control})")
    debug.trace_object(5, upload_control)
    encoded_image = encode_image_file(upload_control.value)
    return encoded_image


def upload_image_file(files):
    """Return path for each of the FILES"""
    debug.trace(4, f"upload_image_file({files})")
    images = [encode_image_file(file.name) for file in files]
    return images


def run_ui(use_img2img=None):
    """Run user interface via gradio serving by default at localhost:7860
    Note: The environment variable GRADIO_SERVER_NAME can be used to serve via 0.0.0.0"""
    import gradio as gr                 # pylint: disable=import-outside-toplevel, redefined-outer-name
    css = """
            .gradio-container {
                font-family: 'IBM Plex Sans', sans-serif;
            }
            .gr-button {
                color: white;
                border-color: black;
                background: black;
            }
            input[type='range'] {
                accent-color: black;
            }
            .dark input[type='range'] {
                accent-color: #dfdfdf;
            }
            .container {
                max-width: 730px;
                margin: auto;
                padding-top: 1.5rem;
            }
            #gallery {
                min-height: 22rem;
                margin-bottom: 15px;
                margin-left: auto;
                margin-right: auto;
                border-bottom-right-radius: .5rem !important;
                border-bottom-left-radius: .5rem !important;
            }
            #gallery>div>.h-full {
                min-height: 20rem;
            }
            .details:hover {
                text-decoration: underline;
            }
            .gr-button {
                white-space: nowrap;
            }
            .gr-button:focus {
                border-color: rgb(147 197 253 / var(--tw-border-opacity));
                outline: none;
                box-shadow: var(--tw-ring-offset-shadow), var(--tw-ring-shadow), var(--tw-shadow, 0 0 #0000);
                --tw-border-opacity: 1;
                --tw-ring-offset-shadow: var(--tw-ring-inset) 0 0 0 var(--tw-ring-offset-width) var(--tw-ring-offset-color);
                --tw-ring-shadow: var(--tw-ring-inset) 0 0 0 calc(3px var(--tw-ring-offset-width)) var(--tw-ring-color);
                --tw-ring-color: rgb(191 219 254 / var(--tw-ring-opacity));
                --tw-ring-opacity: .5;
            }
            #advanced-btn {
                font-size: .7rem !important;
                line-height: 19px;
                margin-top: 12px;
                margin-bottom: 12px;
                padding: 2px 8px;
                border-radius: 14px !important;
            }
            #advanced-options {
                display: none;
                margin-bottom: 20px;
            }
            .footer {
                margin-bottom: 45px;
                margin-top: 35px;
                text-align: center;
                border-bottom: 1px solid #e5e5e5;
            }
            .footer>p {
                font-size: .8rem;
                display: inline-block;
                padding: 0 10px;
                transform: translateY(10px);
                background: white;
            }
            .dark .footer {
                border-color: #303030;
            }
            .dark .footer>p {
                background: #0b0f19;
            }
            .acknowledgments h4{
                margin: 1.25em 0 .25em 0;
                font-weight: bold;
                font-size: 115%;
            }
            .animate-spin {
                animation: spin 1s linear infinite;
            }
            @keyframes spin {
                from {
                    transform: rotate(0deg);
                }
                to {
                    transform: rotate(360deg);
                }
            }
            #share-btn-container {
                display: flex; padding-left: 0.5rem !important; padding-right: 0.5rem !important; background-color: #000000; justify-content: center; align-items: center; border-radius: 9999px !important; width: 13rem;
                margin-top: 10px;
                margin-left: auto;
            }
            #share-btn {
                all: initial; color: #ffffff;font-weight: 600; cursor:pointer; font-family: 'IBM Plex Sans', sans-serif; margin-left: 0.5rem !important; padding-top: 0.25rem !important; padding-bottom: 0.25rem !important;right:0;
            }
            #share-btn * {
                all: unset;
            }
            #share-btn-container div:nth-child(-n+2){
                width: auto !important;
                min-height: 0px !important;
            }
            #share-btn-container .wrap {
                display: none !important;
            }
            
            .gr-form{
                flex: 1 1 50%; border-top-right-radius: 0; border-bottom-right-radius: 0;
            }
            #prompt-container{
                gap: 0;
            }
            #prompt-text-input, #negative-prompt-text-input{padding: .45rem 0.625rem}
            #component-16{border-top-width: 1px!important;margin-top: 1em}
            .image_duplication{position: absolute; width: 100px; left: 50px}
    """
    
    block = gr.Blocks(css=css, title="HF Stable Diffusion gradio UI")

    # note: [prompt, negative, guidance]
    txt2img_examples = [
        [
            'A high tech solarpunk utopia in the Amazon rainforest',
            'low quality',
            9, 
        ],
        [
            'A pikachu fine dining with a view to the Eiffel Tower',
            'low quality',
            9, 
        ],
        [
            'A mecha robot in a favela in expressionist style',
            'low quality, 3d, photorealistic',
            9, 
        ],
        [
            'an insect robot preparing a delicious meal',
            'low quality, illustration',
            9,
        ],
    ]
    
    # note: [input_img, denoise_factor, prompt, negative, guidance ]
    img2img_examples = [
        ## TEST: [ DUMMY_BASE64_IMAGE, 0.75, 'A modern Pacman', 'retro', 7, ],
        [ DUMMY_IMAGE_FILE, 0.75, 'A modern Pacman', 'retro', 7, ],
    ]
    

    # Specify CSS styles and SVG data (TODO3: for what?)
    with block:
        gr.HTML(
            """
                <div style="text-align: center; margin: 0 auto;">
                  <div
                    style="
                      display: inline-flex;
                      align-items: center;
                      gap: 0.8rem;
                      font-size: 1.75rem;
                    "
                  >
                    <svg
                      width="0.65em"
                      height="0.65em"
                      viewBox="0 0 115 115"
                      fill="none"
                      xmlns="http://www.w3.org/2000/svg"
                    >
                      <rect width="23" height="23" fill="white"></rect>
                      <rect y="69" width="23" height="23" fill="white"></rect>
                      <rect x="23" width="23" height="23" fill="#AEAEAE"></rect>
                      <rect x="23" y="69" width="23" height="23" fill="#AEAEAE"></rect>
                      <rect x="46" width="23" height="23" fill="white"></rect>
                      <rect x="46" y="69" width="23" height="23" fill="white"></rect>
                      <rect x="69" width="23" height="23" fill="black"></rect>
                      <rect x="69" y="69" width="23" height="23" fill="black"></rect>
                      <rect x="92" width="23" height="23" fill="#D9D9D9"></rect>
                      <rect x="92" y="69" width="23" height="23" fill="#AEAEAE"></rect>
                      <rect x="115" y="46" width="23" height="23" fill="white"></rect>
                      <rect x="115" y="115" width="23" height="23" fill="white"></rect>
                      <rect x="115" y="69" width="23" height="23" fill="#D9D9D9"></rect>
                      <rect x="92" y="46" width="23" height="23" fill="#AEAEAE"></rect>
                      <rect x="92" y="115" width="23" height="23" fill="#AEAEAE"></rect>
                      <rect x="92" y="69" width="23" height="23" fill="white"></rect>
                      <rect x="69" y="46" width="23" height="23" fill="white"></rect>
                      <rect x="69" y="115" width="23" height="23" fill="white"></rect>
                      <rect x="69" y="69" width="23" height="23" fill="#D9D9D9"></rect>
                      <rect x="46" y="46" width="23" height="23" fill="black"></rect>
                      <rect x="46" y="115" width="23" height="23" fill="black"></rect>
                      <rect x="46" y="69" width="23" height="23" fill="black"></rect>
                      <rect x="23" y="46" width="23" height="23" fill="#D9D9D9"></rect>
                      <rect x="23" y="115" width="23" height="23" fill="#AEAEAE"></rect>
                      <rect x="23" y="69" width="23" height="23" fill="black"></rect>
                    </svg>
                    <h1 style="font-weight: 900; margin-bottom: 7px;margin-top:5px">
                      Stable Diffusion 2.1 Demo
                    </h1>
                  </div>
                  <p style="margin-bottom: 10px; font-size: 94%; line-height: 23px;">
                    Stable Diffusion 2.1 is the latest text-to-image model from StabilityAI. <a style="text-decoration: underline;" href="https://huggingface.co/spaces/stabilityai/stable-diffusion-1">Access Stable Diffusion 1 Space here</a><br>For faster generation and API
                    access you can try
                    <a
                      href="http://beta.dreamstudio.ai/"
                      style="text-decoration: underline;"
                      target="_blank"
                      >DreamStudio Beta</a
                    >.</a>
                  </p>
                </div>
            """
        )

        # Specify the main form
        # TODO3: be consistent in use of xyz_control
        with gr.Group():
            with gr.Box():
                ## TODO: drop 'rounded', border, margin, and other options no longer supported (see log)
                with gr.Row(elem_id="prompt-container").style(mobile_collapse=False, equal_height=True):
                    with gr.Column():
                        prompt_control = gr.Textbox(
                            label="Enter your prompt",
                            show_label=False,
                            max_lines=1,
                            placeholder="Enter your prompt",
                            elem_id="prompt-text-input",
                        ).style(
                            border=(True, False, True, True),
                            rounded=(True, False, False, True),
                            container=False,
                        )
                        negative_control = gr.Textbox(
                            label="Enter your negative prompt",
                            show_label=False,
                            max_lines=1,
                            placeholder="Enter a negative prompt",
                            elem_id="negative-prompt-text-input",
                        ).style(
                            border=(True, False, True, True),
                            rounded=(True, False, False, True),
                            container=False,
                        )
                    btn = gr.Button("Generate image").style(
                        margin=False,
                        rounded=(False, True, True, False),
                        full_width=False,
                    )
    
            gallery = gr.Gallery(
                label="Generated images", show_label=False, elem_id="gallery"
            ).style(grid=[2], height="auto")
    
    
            with gr.Accordion("Advanced settings", open=False):
            #    gr.Markdown("Advanced settings are temporarily unavailable")
            #    samples = gr.Slider(label="Images", minimum=1, maximum=4, value=4, step=1)
            #    steps = gr.Slider(label="Steps", minimum=1, maximum=50, value=45, step=1)
                 guidance_control = gr.Slider(
                    label="Guidance scale", minimum=1, maximum=30, value=GUIDANCE_SCALE, step=0.1
                 )
                 num_control = gr.Slider(
                    label="Number of images", minimum=1, maximum=10, value=2, step=1
                 )
                 img2img_control = gr.Checkbox(label="Use img2img?", value=use_img2img, interactive=True)
                 denoise_control = gr.Slider(label="Denoising factor", minimum=0, maximum=1, value=DENOISING_FACTOR, step=0.05)
                 ## TODO?:
                 input_image_control = gr.Image(label="Input image")  ## TODO?: type='pil'
                 upload_control = gr.UploadButton(label="Upload image", file_types=["image"])
                 interrogate_control = gr.Button(label="CLIP Interrogator")
                 
            #    seed = gr.Slider(
            #        label="Seed",
            #        minimum=0,
            #        maximum=2147483647,
            #        step=1,
            #        randomize=True,
            #    )

            input_controls = [prompt_control, negative_control, guidance_control, num_control]
            output_controls = [gallery]
            infer_fn = infer
            examples = txt2img_examples
            if use_img2img:
                infer_fn = infer_img2img
                input_controls = ([input_image_control, denoise_control] + input_controls)
                examples = img2img_examples
            ex = gr.Examples(examples=examples, fn=infer_fn,
                             inputs=input_controls,
                             outputs=output_controls, cache_examples=False)
            ex.dataset.headers = [""]
            negative_control.submit(infer_fn, inputs=input_controls, outputs=output_controls, postprocess=False)
            prompt_control.submit(infer_fn, inputs=input_controls, outputs=output_controls, postprocess=False)
            btn.click(infer_fn, inputs=input_controls, outputs=output_controls, postprocess=False)
            # TODO1: fix
            upload_control.upload(fn=upload_image_file, inputs=[upload_control], outputs=[input_image_control],
                                  postprocess=False)
            #
            def change_examples():
                """Change examples used in UI if img2img_control checked"""
                debug.trace(4, "change_examples()")
                debug.trace_object(5, img2img_control)
                ex.examples = (img2img_examples if img2img_control.value else txt2img_examples)
            #
            # TODO2: use one listener
            # TODO1: fix
            img2img_control.change(fn=change_examples, inputs=[], outputs=[])
            img2img_control.select(fn=change_examples, inputs=[], outputs=[])
            #
            def use_clip_for_prompt():
                """Run CLIP interrogator over image and send result to prompt field"""
                debug.trace(4, "use_clip_for_prompt()")
                debug.trace_object(5, input_image_control)
                prompt_control.value = infer_img2txt(input_image_control.value)
            #
            # TODO1: fix
            interrogate_control.click(fn=use_clip_for_prompt, inputs=[], outputs=[])
            
            #advanced_button.click(
            #    None,
            #    [],
            #    text,
            #    _js="""
            #    () => {
            #        const options = document.querySelector("body > gradio-app").querySelector("#advanced-options");
            #        options.style.display = ["none", ""].includes(options.style.display) ? "flex" : "none";
            #    }""",
            #)
            gr.HTML(
                """
                    <div class="footer">
                        <p>Model by <a href="https://huggingface.co/stabilityai" style="text-decoration: underline;" target="_blank">StabilityAI</a> - backend running JAX on TPUs due to generous support of <a href="https://sites.research.google/trc/about/" style="text-decoration: underline;" target="_blank">Google TRC program</a> - Gradio Demo by 🤗 Hugging Face
                        </p>
                    </div>
               """
            )
            with gr.Accordion(label="License", open=False):
                gr.HTML(
                    """<div class="acknowledgments">
                        <p><h4>LICENSE</h4>
    The model is licensed with a <a href="https://huggingface.co/stabilityai/stable-diffusion-2/blob/main/LICENSE-MODEL" style="text-decoration: underline;" target="_blank">CreativeML OpenRAIL++</a> license. The authors claim no rights on the outputs you generate, you are free to use them and are accountable for their use which must not go against the provisions set in this license. The license forbids you from sharing any content that violates any laws, produce any harm to a person, disseminate any personal information that would be meant for harm, spread misinformation and target vulnerable groups. For the full list of restrictions please <a href="https://huggingface.co/spaces/CompVis/stable-diffusion-license" target="_blank" style="text-decoration: underline;" target="_blank">read the license</a></p>
                        <p><h4>Biases and content acknowledgment</h4>
    Despite how impressive being able to turn text into image is, beware to the fact that this model may output content that reinforces or exacerbates societal biases, as well as realistic faces, pornography and violence. The model was trained on the <a href="https://laion.ai/blog/laion-5b/" style="text-decoration: underline;" target="_blank">LAION-5B dataset</a>, which scraped non-curated image-text-pairs from the internet (the exception being the removal of illegal content) and is meant for research purposes. You can read more in the <a href="https://huggingface.co/CompVis/stable-diffusion-v1-4" style="text-decoration: underline;" target="_blank">model card</a></p>
                   </div>
                    """
                )
                
    block.queue(concurrency_count=80, max_size=100).launch(max_threads=150)

                
#-------------------------------------------------------------------------------
# Runtime support

def main():
    """Entry point"""

    # Parse command line argument, show usage if --help given
    # TODO? auto_help=False
    main_app = Main(description=__doc__,
                    boolean_options=[(BATCH_ARG, "Use batch mode--no UI"),
                                     (SERVER_ARG, "Run flask server"),
                                     (UI_ARG, "Show user interface"),
                                     (TXT2IMG_ARG, "Run text-to-image--the default"),
                                     (IMG2IMG_ARG, "Run image-to-image"),
                                     (IMG2TXT_ARG, "Run image-to-text: clip interrogator")],
                    text_options=[(PROMPT_ARG, "Positive prompt"),
                                  (NEGATIVE_ARG, "Negative prompt"),
                                  ],
                    int_options=[(GUIDANCE_ARG, GUIDANCE_HELP)],
                    float_options=[(DENOISING_ARG, "Denoising factor for img2img")])
    debug.trace_object(5, main_app)
    debug.assertion(main_app.parsed_args)
    #
    input_image_file = main_app.filename
    BATCH_MODE_DEFAULT = (input_image_file != "-")
    batch_mode = main_app.get_parsed_option(BATCH_ARG, BATCH_MODE_DEFAULT)
    server_mode = main_app.get_parsed_option(SERVER_ARG)
    ui_mode = main_app.get_parsed_option(UI_ARG)
    prompt = main_app.get_parsed_option(PROMPT_ARG, PROMPT)
    negative_prompt = main_app.get_parsed_option(NEGATIVE_ARG, NEGATIVE_PROMPT)
    guidance = main_app.get_parsed_option(GUIDANCE_ARG, GUIDANCE_SCALE)
    use_img2img = main_app.get_parsed_option(IMG2IMG_ARG, USE_IMG2IMG)
    use_img2txt = main_app.get_parsed_option(IMG2TXT_ARG, USE_IMG2TXT)
    use_txt2img = main_app.get_parsed_option(TXT2IMG_ARG, not (use_img2img or use_img2txt))
    denoising_factor = main_app.get_parsed_option(DENOISING_ARG)
    ## TODO?:
    debug.assertion((use_txt2img ^ use_img2img) or use_img2txt)
    # TODO2: BASENAME and NUM_IMAGES (options)
    ## TODO: x_mode = main_app.get_parsed_option(X_ARG)
    debug.assertion(not (batch_mode and server_mode))

    # Invoke UI via HTTP unless in batch or server mode
    show_gpu_usage()
    init_stable_diffusion()
    if batch_mode:
        # Optionally convert input image into base64
        b64_image_encoding = (encode_image_file(input_image_file) if (use_img2img or use_img2txt) else None)

        # Run image generation (or text from image)
        if use_img2txt:
            description = infer_img2txt(b64_image_encoding)
            print(description)
        else:
            images = (infer(prompt, negative_prompt, guidance, skip_img_spec=True) if (not use_img2img)
                      else infer_img2img(b64_image_encoding, denoising_factor, prompt, negative_prompt, guidance, skip_img_spec=True))
            # Save result to disk
            for i, image_encoding in enumerate(images):
                write_image_file(f"{BASENAME}-{i + 1}.png", image_encoding)
            # TODO2: get list of files via infer()
            file_spec = " ".join(gh.get_matching_files(f"{BASENAME}*png"))  
            print(f"See {file_spec} for output image(s).")
    # Start restful server
    elif server_mode:
        debug.assertion(SD_URL)
        debug.assertion(not sd_instance.server_url)
        debug.trace_object(5, flask_app)
        flask_app.run(host=SD_URL, port=SD_PORT, debug=SD_DEBUG)
    # Start UI
    elif ui_mode:
        debug.assertion(main_app.filename == "-")
        run_ui(use_img2img=use_img2img)
    # Otherwise, show command-line options
    else:
        ## TODO3: expose print_usage directly through main_app
        main_app.parser.print_usage()
    show_gpu_usage()


if __name__ == '__main__':
    debug.trace_current_context(level=TL.QUITE_DETAILED)
    main()



#! /usr/bin/env python3
# 
# Similar to Unix cut command but with support for CSV files. Also modelled
# after perl script with support for treating runs of whitespace as tab.
#
# Notes:
# - Input processing based on csv example (seee https://docs.python.org/2/library/csv.html).
# - The bulk of the work is parsing the field specification After that,
#   the processing is simple row readidng and column extraction (i.e., subsetting).
# - Have option for setting delim to tab to avoid awkward spec under bash (e.g., --output-delim $'\t').
# - The CSV dialect defaults to Excel as with csv module (see csv.py).
# - Warning: by default quotes are added to all values --csv output (a la QUOTE_ALL) unless Excel dialect used.
#

#
# TODO:
# - ** Make --csv output default to pyspark dialect.
# - * Isolate csv support from Script class.
# - Add option for specifying output dialect.
# - Add support for selecting by columns instead of fields (e.g., as with -c1-40 with cut command).
#

"""Extracts columns from a file as with Unix cut command"""

# Standard modules
import argparse
import csv
import re
import sys

# Installed modules
## OLD: import more_itertools
import functools
import operator

# Local modules
from mezcla import data_utils as du
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla.main import Main
from mezcla.my_regex import my_re
from mezcla import system

# Fill out constants for switches omitting leading dashes (e.g., DEBUG_MODE = "debug-mode")
FIELDS = "fields"                       # field indices (1-based or symbolic)
F_OPT = "f"                             # alias for fields
EXCLUDE_OPT = "exclude"                 # fields to exclude (1-based or symbolic)
X_OPT = "x"                             # alias for exclude
ENCODE_OPT = "encode"                   # use repr for newlines, etc.
FIX = "fix"                             # convert runs of spaces into a tab
CSV = "csv"                             # comma-separated value format
TSV = "tsv"                             # tab-separated value format
OUTPUT_CSV = "output-csv"               # CSV for output
OUTPUT_TSV = "output-tsv"               # TSV for output
CONVERT_DELIM = "convert-delim"         # convert input csv to tsv (or vice versa)
SNIFFER_ARG = "sniffer"                 # run CSV sniffer to detect dialect
## TODO
## INPUT_CSV = "input-csv"
## OUTPUT_CSV = "output-csv"
## INPUT_TSV = "input-tsv"
## OUTPUT_TSV = "output-tsv"
DELIM = "delim"                         # input delimiter
OUT_DELIM = "output-delim"              # output delimiter if not same for input
ALL_FIELDS = "all-fields"               # use all fields in output (e.g., for delimiter conversion)
TAB = "\t"
SPACE = " "
COMMA = ","
## TODO: make -style suffix optional (e.g., --tab[-style])
EXCEL_STYLE = "excel-style"             # use Excel dialect for CSV (redundant with default)
UNIX_STYLE = "unix-style"               # use Unix dialect for CSV (otherwise Excel is used)
PYSPARK_STYLE = "pyspark-style"         # use PySpark dialect for CSV (otherwise Unix is used)
## TEMP: workaround problem under MacOS
FORCE_SINGLE_LINE = system.getenv_bool(
    "FORCE_SINGLE_LINE", False,
    description="Force single line quotedchar handling (i.e., non-strict mode)")
TAB_STYLE = "tab-style"                 # use (non-Excel) tab dialect
DIALECT = "dialect"                     # --dialect option
OUTPUT_DIALECT = "output-dialect"       # --output-dialect option
EXCEL_DIALECT = "excel"                 # dialect parameter for Excel style
UNIX_DIALECT = "unix"                   # "" for Unix style
PYSPARK_DIALECT = "pyspark"             # "" for Pyspark style
SNIFFER_LOOKAHEAD = 65536               # buffer size for guessing dialect (64k)
TAB_DIALECT = "tab"                     # dialect option for TSV
SINGLE_LINE = "single-line"             # collapse multi-line fields into one
MAX_FIELD_LEN = "max-field-len"         # value length before elided
## TODO: TODO_ARG = "TODO-arg"          # TODO: comment
NEW_FIX = system.getenv_bool(
    "NEW_FIX", False,
    desc="HACK: Fix for --fix bug for whitespace to tabs")
NUM_FN_SHORTCUTS = 9
CSV_FORMAT = system.getenv_bool(
    "CSV_FORMAT", False,
    desc="Use CSV instead of TSV")
MAX_FIELD_SIZE = system.getenv_int(
    "MAX_FIELD_SIZE", -1,
    desc="Overide for default max field size (128k)")

#...............................................................................

MAX_VALUE_LEN = 128
#
def elide_values(in_list, max_len=MAX_VALUE_LEN):
    """Elide each of the values in IN_LIST (up to MAX_LEN each).
    Note: Returns list of strings"""
    # EX: elide_values(["1234567890", 1234567890, True, False], max_len=4) => ["1234...", "1234...", "True", "Fals..."]
    new_list = []
    for item in in_list:
        new_list.append(gh.elide(system.to_text(item), max_len))
    debug.trace_fmt(7, "elide_values({l}, [{m}]) => {r}", l=in_list, m=max_len, r=new_list)
    return new_list

def flatten_list_of_strings(list_of_str):
    """Flatten out LIST_OF_STR"""
    # EX: flatten_list_of_strings([["l1i1", "l1i2"], ["l2i1"]]) => ["l1i1", "l1i2", "l2i1"]
    result = functools.reduce(operator.concat, list_of_str)
    debug.trace(5, f"flatten_list_of_strings({list_of_str}) => {result}")
    return result

#...............................................................................
# TODO: Put following in separate module (e.g., data_utils.py)
#
# Note:
#
# - Dialect definitions from source for the standard CSV module (csv.py)
#
#   class excel(Dialect):
#       """Describe the usual properties of Excel-generated CSV files."""
#       delimiter = ','
#       quotechar = '"'
#       doublequote = True                     [ " => "" if embedded ]
#       skipinitialspace = False
#       lineterminator = '\r\n'
#       quoting = QUOTE_MINIMAL
#   register_dialect("excel", excel)
#   
#   class unix_dialect(Dialect):
#       """Describe the usual properties of Unix-generated CSV files."""
#       delimiter = ','
#       quotechar = '"'
#       doublequote = True
#       skipinitialspace = False
#       lineterminator = '\n'
#       quoting = QUOTE_ALL
#   register_dialect("unix", unix_dialect)
# 
# - Both double a double quote if embedded (e.g., "I said ""Hey!"" then.")
#

class pyspark_dialect(csv.Dialect):
    """CSV module dialect for Pyspark CSV files."""
    delimiter = ','
    quotechar = '"'
    doublequote = False          # uses escaped double quote when embedded
    escapechar = '\\'
    skipinitialspace = False
    lineterminator = '\n'
    quoting = csv.QUOTE_MINIMAL  # only delimiter, double quote or end-of-line
#
csv.register_dialect("pyspark", pyspark_dialect)
#
# note: Uses hive as alias for pyspark.
csv.register_dialect("hive", pyspark_dialect)

class tab_dialect(csv.Dialect):
    """TSV module dialect for tab-separated values (non-Excel)."""
    delimiter = TAB
    ## OLD: quotechar = ''               # default of '"' leads to multiline rows
    quotechar = ('"' if not FORCE_SINGLE_LINE else '')
    doublequote = False          # uses escaped double quote when embedded
    # TODO: use special Unicode space-like character
    escapechar = '\\'
    skipinitialspace = False     # keep leaing space afer delimiter
    lineterminator = '\n'
    quoting = csv.QUOTE_NONE     # no special processing for quotes
#
csv.register_dialect("tab", tab_dialect)

#...............................................................................

class Script(Main):
    """Input processing class"""
    inclusion_spec = ''
    exclusion_spec = ''
    fields = []
    exclude_fields = []
    encode_values = False
    fix = False
    ## OLD: delimiter = TAB
    delimiter = None
    output_delimiter = None
    csv = CSV_FORMAT
    csv_reader = None
    all_fields = False
    dialect = None
    output_dialect = None
    run_sniffer = False
    single_line = False
    max_field_len = None
    ## TODO: todo_arg = ...

    def setup(self):
        """Check results of command line processing"""
        # Note: The defaults above might be changed based on other arguments;
        # but, the user-specified value is normally not overridden.
        debug.trace_fmtd(5, "Script.setup(): self={s}", s=self)

        # Check main options
        # note: several options regarding fields including --Fn aliases
        ## OLD: fields = self.get_parsed_option(FIELDS, self.fields)
        for i in range(NUM_FN_SHORTCUTS):
            if self.get_parsed_option(f"F{i + 1}"):
                self.fields.append(str(i + 1))
        FIELDS_DEFAULT = ",".join(self.fields)
        debug.trace_expr(4, FIELDS_DEFAULT)
        inclusion_spec = self.get_parsed_option(F_OPT, FIELDS_DEFAULT)
        self.inclusion_spec = self.get_parsed_option(FIELDS, inclusion_spec)
        EXCLUDE_DEFAULT = ",".join(self.exclude_fields)
        exclusion_spec = self.get_parsed_option(F_OPT, EXCLUDE_DEFAULT)
        self.exclusion_spec = self.get_parsed_option(EXCLUDE_OPT, exclusion_spec)
        debug.assertion(not (self.inclusion_spec and self.exclusion_spec))
        self.encode_values = self.get_parsed_option(ENCODE_OPT, self.encode_values)
        ## NOTES: Fields initialization postponed until columns reads to allow ffor symbolic names
        ## OLD:
        ## if fields:
        ##     self.fields = self.parse_field_spec(fields)
        self.all_fields = self.get_parsed_option(ALL_FIELDS, (not self.fields))
        debug.assertion(not (self.fields and self.all_fields))
        #
        self.fix = self.get_parsed_option(FIX, self.fix)

        # Check delimiter options
        # Note: output delimiter defaults to input unless convert specified
        self.csv = self.get_parsed_option(CSV, self.csv)
        if self.csv:
            self.delimiter = COMMA
        tsv = self.get_parsed_option(TSV, not self.csv)
        if tsv:
            self.delimiter = TAB
        self.delimiter = self.get_parsed_option(DELIM, self.delimiter)
        #
        debug.assertion(self.output_delimiter is None)
        if self.get_parsed_option(OUTPUT_CSV):
            self.output_delimiter = COMMA
        if self.get_parsed_option(OUTPUT_TSV):
            self.output_delimiter = TAB
        if self.output_delimiter is None:
            self.output_delimiter = self.delimiter
        if self.get_parsed_option(CONVERT_DELIM):
            self.output_delimiter = TAB if (self.delimiter == COMMA) else COMMA
        if (self.output_delimiter is None):
            self.output_delimiter = self.delimiter
        #
        self.run_sniffer = self.get_parsed_option(SNIFFER_ARG, self.run_sniffer)
        debug.assertion(not (self.csv and (self.delimiter != COMMA)))
        self.output_delimiter = self.get_parsed_option(OUT_DELIM, (self.output_delimiter or self.delimiter))
        self.single_line = self.get_parsed_option(SINGLE_LINE, self.single_line)
        self.max_field_len = self.get_parsed_option(MAX_FIELD_LEN, self.max_field_len)
        # self.todo_arg = self.get_parsed_option(TODO_ARG, self.todo_arg)

        # Check CSV dialet options
        debug.assertion(system.just_one_non_null([self.get_parsed_option(o, None) for o in 
                                                  [DIALECT, EXCEL_DIALECT, UNIX_DIALECT, PYSPARK_DIALECT, TAB_DIALECT]]))
        pyspark_style = self.get_parsed_option(PYSPARK_STYLE)
        if self.get_parsed_option(EXCEL_STYLE):
            self.dialect = EXCEL_DIALECT
        elif pyspark_style:
            self.dialect = PYSPARK_DIALECT
        elif (self.get_parsed_option(UNIX_STYLE) or (self.delimiter == COMMA)):
            self.dialect = UNIX_DIALECT
        elif (self.get_parsed_option(TAB_STYLE) or (self.delimiter == TAB)):
            self.dialect = TAB_DIALECT
        self.dialect = self.get_parsed_option(DIALECT, self.dialect)
        default_output_dialect = (self.output_dialect or self.dialect) if not self.run_sniffer else self.output_dialect
        if (self.output_delimiter == TAB):
            default_output_dialect = TAB_DIALECT
        if (self.output_delimiter == COMMA):
            default_output_dialect = PYSPARK_DIALECT if pyspark_style else UNIX_DIALECT
        self.output_dialect = self.get_parsed_option(OUTPUT_DIALECT, default_output_dialect)
        # TODO: see if there is an option to determine number of fields (before reading data)
        ## if self.all_fields:
        ##    self.fields = range(self.csv_reader.num_fields)

        # Trace instane
        debug.trace_object(5, self, label="Script instance")
        return

    def parse_field_spec(self, field_spec, columns):
        """Convert the FIELD_SPEC from string to list of integers. The specification can contain numeric ranges (e.g., "3-5") or comma-separated values (e.g., "7,9,11").
        Symbolic names can also be used (e.g., "sepal_width,petal_width")
        Note: throws exception if fields are not integers"""
        ## TODO3: decompose using helper functions
        # EX: self.parse_field_spec("1,2,5-8,3,4") => [1, 2, 5, 6, 7, 8, 3, 4]
        debug.trace(5, f"parse_field_spec({field_spec}, {columns}); self={self}")

        # Convert labels to 1-based offsets (TODO3: find CSV spec and clarify label chars)
        # NOTE: dashes in field names not supported (TODO2: subsitute non-ascii dash)
        in_field_spec = field_spec
        if my_re.search(r"[A-Za-z]", field_spec):
            ## TODO4: quote fields that should be used as is; add alternative to dash
            debug.trace(4, "FYI: Fieldnames currently are Python-like Ascii identifiers")
        while my_re.search(r"([a-z][a-z0-9_]+)", field_spec, flags=my_re.IGNORECASE):
            field_name = my_re.group(1)
            if field_name not in columns:
                system.print_error(f"Error: Unable to resolve field {field_name!r}: {in_field_spec=}")
                if field_name.lower() in str(field_spec).lower():
                    debug.trace(4, "Try matching case")
                break
            field_pos = (1 + columns.index(field_name))
            field_spec = my_re.pre_match() + str(field_pos) + my_re.post_match()
            debug.trace(4, f"Replaced {field_name!r} with {field_pos}")
        
        # Normalize the field specification
        debug.assertion(not re.search(r"[0-9] [0-9]", field_spec))
        field_spec = field_spec.replace(SPACE, "")
        debug.assertion(",," not in field_spec)
        field_spec = re.sub(r",,+", ",", field_spec)
        debug.assertion(not (field_spec.startswith(",") or field_spec.endswith(",")))
        field_spec = re.sub(r"(^,)|(,$)", "", field_spec)
        debug.trace_fmtd(5, "normalized field_spec: {fs}", fs=field_spec)

        # Replace ranges with comma-separated cols (e.g., "5-7" => "5,6,7")
        if "-" in field_spec:
            # Supply missing first and last column (e.g., "^-3" => "1-3" and "2-$" => "2-3")
            field_spec = my_re.sub(r"^\-", "1-", field_spec)
            field_spec = my_re.sub(r"\-$", f"-{len(columns)}", field_spec)

            # Replace ranges
            while my_re.search(r"([0-9]+)\-([0-9]+)", field_spec):
                range_spec = my_re.group(0)
                debug.trace_fmtd(4, "Converting range: {r}", r=range_spec)
                start = int(my_re.group(1))
                end = int(my_re.group(2))
                debug.assertion(start < end)

                # Add comma-separated values in range
                subfield_spec = ""
                while (start <= end):
                    subfield_spec += str(start) + ","
                    start += 1
                if subfield_spec.endswith(","):
                    subfield_spec = subfield_spec[:-1]

                # Replace the range specification with the delimited values
                debug.trace_fmtd(5, "subfield_spec: {ss}", ss=subfield_spec)
                field_spec = field_spec.replace(range_spec, subfield_spec, 1)
        debug.trace_fmtd(5, "expanded field_spec: {fs}", fs=field_spec)

        # Convert text field specification into list
        field_list = []
        if field_spec:
            field_list = [int(f) for f in field_spec.split(",")]
        debug.assertion(field_list)
        debug.trace_fmtd(4, "parse_field_spec() => {fl}", fl=field_list)
        return field_list
    
    def run_main_step(self):
        """Main processing step: read each line (i.e. row) and extract specified columns.
        Note: The fields are 1-based (i.e., first column specified 1 not 0)"""
        debug.trace_fmtd(4, "run_main_step()")

        # Overide the maxium field size if specified
        if MAX_FIELD_SIZE > -1:
            old_limit = csv.field_size_limit()
            debug.assertion(MAX_FIELD_SIZE > old_limit)
            csv.field_size_limit(MAX_FIELD_SIZE)
            debug.trace(4, f"Set max field size to {MAX_FIELD_SIZE}; was {old_limit}")
        
        # If unspecified, try to determine dialect for CSV input automatically
        # Notes: doesn't recognize escapes properly, such as \"); and,
        # sniffer doesn't work for stdin due to python limitation with backtracking.
        if ((self.delimiter == COMMA) and (not self.dialect) and self.run_sniffer):
            debug.assertion(self.input_stream != sys.stdin)
            self.dialect = csv.Sniffer().sniff(self.input_stream.read(SNIFFER_LOOKAHEAD))
            debug.trace_object(4, self.dialect, "csv sniffer")
            self.input_stream.seek(0)
            if (self.output_dialect is None):
                self.output_dialect = self.dialect

        # Optionally, fixup input if TSV changing multiple spaces into single tab.
        # Note: makes pass through data, writes to temp file, and then resets input stream to
        # read from the temp file.
        # TODO: have option to distinguish old-style loose fix (3+ spaces) from strict fix (all whitespace)
        if self.fix:
            debug.assertion(self.delimiter == TAB)
            temp_file_stream = system.open_file(self.temp_file, mode="w")
            num_fixed = 0
            for line in self.input_stream.readlines():
                ## OLD: line = re.sub(r"   *", TAB, line)
                new_line = re.sub(r" +", TAB, line)
                if (new_line != line):
                    num_fixed += 1
                    line = new_line
                debug.assertion(SPACE not in line)
                temp_file_stream.write(line)
            debug.trace(4, f"Fixed {num_fixed} lines")
            ## OLD: self.input_stream.seek(0)
            temp_file_stream.close()
            self.input_stream = system.open_file(self.temp_file, mode="r")
            ## HACK: pretend reading from stdin
            sys.stdin = self.input_stream

        # Create reader and writer
        debug.trace_expr(4, self.delimiter, self.output_delimiter, self.dialect, self.output_dialect)
        ## BAD: self.csv_reader = csv.reader(iter(system.stdin_reader()), delimiter=self.delimiter, quotechar='"')
        if (self.input_stream != sys.stdin):
            # note: silly csv.reader requirement for newline option to open (TODO, open what?)
            # TODO: add support for multiple filenames
            self.input_stream = system.open_file(self.filename, newline="")
            debug.assertion(not self.other_filenames)
        self.csv_reader = csv.reader(self.input_stream, delimiter=self.delimiter, 
                                     dialect=self.dialect)
        csv_writer = csv.writer(sys.stdout, delimiter=self.output_delimiter, 
                                dialect=self.output_dialect)
        debug.trace_object(5, self.csv_reader, "csv_reader")
        debug.trace_object(5, self.csv_reader.dialect, "csv_reader.dialect")
        debug.trace_object(5, csv_writer, "csv_writer")
        debug.trace_object(5, csv_writer.dialect, "csv_writer.dialect")

        # Iterate through the rows, outputting subset of columns
        last_row_length = None
        num_rows = 0
        num_cols = None
        columns = None
        for i, row in enumerate(self.csv_reader):
            if NEW_FIX and (self.delimiter == TAB):
                # Strip leading spaces and replace other multiple spaces by a tab
                debug.trace_fmt(7, "old R{n}: {r}", n=(i + 1), r=row)
                ## OLD: row = list(more_itertools.flatten([re.sub(" +", TAB, f) for f in row]))
                ## OLD: row = flatten_list_of_strings([re.split(r" +", f) for f in row])
                line = TAB.join(row)
                line = re.sub("^ +", "", line)
                line = re.sub(" +", TAB, line)
                row = line.split(TAB)
                debug.assertion(not any(SPACE in field for field in row))
            if i == 0:
                # note: Byte order mark (BOM) is removed
                # TODO3: warn about delimiter mismatch
                columns = row
                BOM = '\ufeff'
                if columns and columns[0].startswith(BOM):
                    columns[0] = columns[0][len(BOM):]
                if self.inclusion_spec:
                    self.fields = self.parse_field_spec(self.inclusion_spec, columns)
                if self.exclusion_spec:
                    self.exclude_fields = self.parse_field_spec(self.exclusion_spec, columns)
            debug.trace_fmt(6, "R{n}: {r}", n=(i + 1), r=row)
            debug.trace_fmt(5, "R{n}: len(row)={l} [{rspec}]", n=(i + 1), l=len(row), rspec=elide_values(row))
            debug.assertion((len(row) == last_row_length) or (not last_row_length))
            last_row_length = len(row)
            # TODO: rework i references in terms of num_rows
            num_rows += 1
            if num_cols is None:
                num_cols = last_row_length

            # Derive the fields to extract if all to be extracted
            debug.trace_fmt(7, "pre f={f} all={a}", f=self.fields, a=self.all_fields)
            if ((not self.fields) and self.all_fields):
                ## OLD: self.fields = [(c + 1) for c in range(len(row))]
                self.fields = [(c + 1) for c in range(len(row)) if (c + 1) not in self.exclude_fields]
                if not self.fields:
                    ## OLD: system.print_stderr("Error: No items in header row at line {l}", l=(i + 1))
                    system.print_stderr("Error: No items in row at line {l}", l=(i + 1))
            debug.trace_fmt(7, "post f={f}", f=self.fields)
            debug.assertion(self.fields)

            # Output line with fields joined by (output) separator
            ## OLD:
            ## line = ""
            ## for f in self.fields:
            ##     valid_field_number = (1 <= f <= len(row))
            ##     debug.assertion(valid_field_number)
            ##     ## OLD: line += (row[f - 1] if (1 <= f <= len(row)) else "")
            ##     line += (row[f - 1] if valid_field_number else "")
            ##     line += self.output_delimiter
            ## if line.endswith(self.output_delimiter):
            ##    line = line[:-1]
            ## debug.trace_fmt(8, "line={l}", l=line)
            ## print(line)
            ##
            output_row = []
            for f in self.fields:
                valid_field_number = (1 <= f <= len(row))
                debug.trace_expr(5, f)
                debug.assertion(valid_field_number, f"field {f}")
                ## OLD: output_row.append(row[f - 1] if valid_field_number else "")
                column = row[f - 1] if valid_field_number else ""
                if self.single_line:
                    column = re.sub(r"\s", SPACE, column)
                if self.max_field_len:
                    column = gh.elide(column, max_len=self.max_field_len)
                if self.encode_values:
                    ## TODO4: maxcount of 1 for left and right
                    column = repr(column).strip("'")
                output_row.append(column)
            debug.trace_expr(6, output_row)
            try:
                csv_writer.writerow(output_row)
            except:
                ## TODO2: add ignore-errors option
                system.print_exception_info("row output")

        # Do sanity checks
        # Note: this compares row extraction against Pandas dataframe
        ## OLD: if (self.input_stream != sys.stdin):
        if (debug.debugging() and (self.input_stream != sys.stdin)):
            debug.trace(4, "note: csv vs. pandas row count sanity check")
            ## BAD: debug.assertion(num_rows == len(du.read_csv(self.filename, delimiter=self.delimiter))
            dataframe = du.read_csv(self.filename, delimiter=self.delimiter, dialect=self.dialect)
            valid_dataframe = (dataframe is not None)
            debug.assertion(valid_dataframe)
            if valid_dataframe:
                df_num_rows = 1 + len(dataframe)
                df_num_cols = len(dataframe.columns)
                debug.trace_fmt(4, "csv dimensions: {nr}x{nc}", nr=num_rows, nc=num_cols)
                debug.trace_fmt(4, "pandas dimensions: {nr}x{nc}", nr=df_num_rows, nc=df_num_cols)
                debug.assertion(num_rows == df_num_rows)
                debug.assertion(num_cols == df_num_cols)

        return

if __name__ == '__main__':
    debug.trace_current_context()
    debug.trace_fmt(4, "Environment options: {eo}",
                    eo=system.formatted_environment_option_descriptions())
    app = Script(
        description=__doc__,
        skip_input=False,
        manual_input=True,
        multiple_files=True,
        boolean_options=(
            [("Fn", "alias for --field n")] +
            [(f"F{i + 1}", argparse.SUPPRESS) for i in range(NUM_FN_SHORTCUTS)] +
            [(CSV, "Comma-separated values ({xls} as per csv module)".format(xls=EXCEL_STYLE)),
             (TSV, "Tab-separated values"),
             OUTPUT_CSV, OUTPUT_TSV,
             (CONVERT_DELIM, "Convert csv to tsv (or vice versa)"),
             (SNIFFER_ARG, "Detect csv dialect by lookahead (file-input only)"),
             ## TODO: INPUT_CSV, OUTPUT_CSV, INPUT_TSV, OUTPUT_TSV,
             (FIX, "Fix up sloppy input (e.g., multiple spaces into tab)--csv fixup not yet supported"),
             (ALL_FIELDS, "Alternative to {f} option".format(f=FIELDS)),
             (EXCEL_STYLE, "Use Excel conventions for CSV files (see csv python package docs)"),
             (PYSPARK_STYLE, "Use PySpark conventions for CSV files (see {f} source)".format(f=__file__)),
             (SINGLE_LINE, "Remove embedded newlines from mult-line fields"),
             (TAB_STYLE, "Non-excel TSV conventions (default)"),
             ## (TODO_ARG, "TODO: arg desc").
             (UNIX_STYLE, "Use Unix conventions for CSV files (see csv python package docs)"),
             (ENCODE_OPT, "Output field encoded via repr (i.e., canonical representation)"),
             ]),
        int_options = [(MAX_FIELD_LEN, "Maximum length per field")],
        text_options=[(DELIM, "Input field separator"),
                      (DIALECT, "CSV module dialect: standard (i.e., excel, excel-tab, or unix) or adhoc (e.g., pyspark, hive)"),
                      (OUTPUT_DIALECT, "dialect for output--defaults to input one"),
                      (FIELDS, "Field specification (1-based or label): single column, range of columns, or comma-separated columns"),
                      (F_OPT, "Alias for --fields"),
                      (OUT_DELIM, "Output field separator"),
                      (EXCLUDE_OPT, "Field specification (1-based or label): single column, range of columns, or comma-separated columns"),
                      (X_OPT, "Alias for --exclude"),
                      ])
    app.run()



#! /usr/bin/env python3
#
# Based on following:
# - Jason Brownlee's blog:
#   https://machinelearningmastery.com/feature-importance-and-feature-selection-with-xgboost-in-python
# - Earlier sklearn scripts:
#   ex: pandas_sklearn.py
#
# Note: another buggy blog.
#

"""Feature Importance and Feature Selection With XGBoost"""

# Standard packages
import re

# Installed packages
# use feature importance for feature selection, with fix for xgboost 1.0.2
## OLD: from numpy import loadtxt
from matplotlib import pyplot as plt
import numpy as np
import xgboost
from xgboost import XGBClassifier
from xgboost import plot_importance
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score
from sklearn.feature_selection import SelectFromModel

# Local packages
from mezcla import data_utils as du
from mezcla import debug
from mezcla import glue_helpers as gh
## OLD
## from mezcla.main import Main
## from mezcla.main import dummy_app
from mezcla import system
from mezcla.system import getenv_bool, getenv_float, getenv_int, getenv_text, getenv_value

# Runtime options
USE_WRAPPER_DEFAULT = re.search(r"^1.0", getattr(xgboost, "__version__", "???"))
USE_WRAPPER = getenv_bool("USE_WRAPPER", USE_WRAPPER_DEFAULT,
                          "Use wrapper for XGBClassifier to work around old bug")
USE_LOADTXT = getenv_bool("USE_LOADTXT", False,
                          "Use Numpy's loadtxt to read data")
USE_DATAFRAME = getenv_bool("USE_DATAFRAME", not USE_LOADTXT,
                            "Retain data in Pandas dataframe format")
SKLEARN_PLOT = getenv_bool("SKLEARN_PLOT", False,
                           "Output sklearn-based plot")
XGBOOST_PLOT = getenv_bool("XGBOOST_PLOT", False,
                           "Use Xgboost version of plot")
INTERACTIVE_PLOT =  getenv_bool("INTERACTIVE_PLOT", False,
                                "Show interactive plot")
DEFAULT_DATAFILE = getenv_text("DATA_FILE", "pima-indians-diabetes.csv")
DEFAULT_OUTPUT_BASE = getenv_value("OUTPUT_BASE", "")
RANDOM_SEED = getenv_int("RANDOM_SEED", 7,
                         "Seed for random number generator")
TEST_PERCENT = getenv_float("TEST_PERCENT", 0.33,
                            "Percentage (n.b., ratio) of data for test set")
DUMP_MODEL = getenv_bool("DUMP_MODEL", False,
                         "Dump model tree(s) with split statistics")
FORCE_FLOAT = getenv_bool("FORCE_FLOAT", False,
                          "Force input daya to be floating point")

# define custom class to fix bug in xgboost 1.0.2
class MyXGBClassifier(XGBClassifier):
    """wrapper class"""
    
    @property
    def coef_(self):
        """coefficient for feature importance"""
        ## BAD: return None
        return self.feature_importances_

def main():
    """Entry point"""
    data_filename = 'pima-indians-diabetes.csv'

    # Check command line arguments
    args = system.get_args()
    show_help = ((len(args) > 1) and (args[1] == "--help"))
    show_usage = ((len(args) > 1) and (args[1] == "--usage"))
    if ((len(args) <= 1) or show_help or show_usage):
        script = gh.basename(args[0])
        system.print_stderr("Usage: {scr} [--help | --usage] [data-file | -]", scr=script)
        system.print_stderr("")
        system.print_stderr("Notes:")
        system.print_stderr("- Use - for data-file to use default.")
        system.print_stderr("- Where data-file is in CSV format (or TSV)")
        ## TODO: show_details = VERBOSE or debug.debugging()
        debugging = debug.detailed_debugging()
        if (show_help or debugging):
            ## TODO: add legend
            system.print_stderr("- Environment options:")
            system.print_stderr("\t" + system.formatted_environment_option_descriptions(include_all=debugging))
        system.exit("")
    
    if ((len(args) > 1) and (not args[1].startswith("-"))):
        data_filename = args[1]
    debug.trace(4, f"data_file={data_filename}")
    if data_filename is None:
        data_filename = DEFAULT_DATAFILE
    output_base = DEFAULT_OUTPUT_BASE
    if not output_base:
        output_base = system.remove_extension(data_filename)

    # load data
    ## OLD: data_path = gh.resolve_path('pima-indians-diabetes.csv')
    ## TODO: strip comment line
    data_path = gh.resolve_path(data_filename)
    ## if re.search(r"[A-Za-z]", gh.run("head -1 {data_path}")):
    ##     debug.trace(4, f"Note: stripping header from {data_path}")
    ##     ## dummy_app = Main([])
    ##     gh.run(gh.run("tail --lines=+2 {data_path} > {dummy_app.temp_file}"))
    ##     data_path = dummy_app.temp_file
    ## OLD: dataset = loadtxt(data_path, delimiter=",")
    if USE_LOADTXT:
        dataset = np.loadtxt(data_path, delimiter=",", skiprows=1)
        labels = [f"_F{i + 1}" for i in range(dataset.shape[1])]
    else:
        in_dtype = None if not FORCE_FLOAT else np.float64
        dataset = du.read_csv(data_path, delimiter=",", dtype=in_dtype)
        labels = list(dataset.columns)
        if not USE_DATAFRAME:
            dataset = dataset.values    # pylint: disable=no-member

    class_col = (len(labels) - 1)
    class_var = labels[class_col]
    feature_labels = labels[0: class_col]
    debug.trace_expr(5, dataset, labels, class_col, delim='\n')

    # split data into X and y
    if USE_DATAFRAME:
        X = dataset[feature_labels]
        Y = dataset[[class_var]]
    else:
        X = dataset[:, 0: class_col]
        Y = dataset[:, class_col]
    #
    # split data into train and test sets
    debug.assertion(0 < TEST_PERCENT < 1)
    X_train, X_test, y_train, y_test = train_test_split(
        X, Y, test_size=TEST_PERCENT, random_state=RANDOM_SEED)
    debug.trace_expr(6, X_train, X_test, y_train, y_test)
    
    # Fit model on all training data proper
    classifier = (XGBClassifier if (not USE_WRAPPER) else MyXGBClassifier)
    ## OK:
    model = classifier()
    ## TEST: model = classifier(feature_names=labels)
    ## OK:
    model.fit(X_train, y_train)
    ## TEST:
    ## dtrain = xgboost.DMatrix(X_train, label=y_train, feature_names=feature_labels)
    ## model = xgboost.train({}, dtrain)
    debug.trace_object(6, model, "model")

    # Optionally dump out the tree(s) in the model
    if DUMP_MODEL:
        model.get_booster().dump_model(f"{output_base}-model-dump.list",
                                       with_stats=True)

    # Optionally plot the feature importance values
    if SKLEARN_PLOT:
        plt.bar(range(len(model.feature_importances_)),
                model.feature_importances_, tick_label=feature_labels)
        plt.title("Feature importances")
        plt.figtext(0, 0, "dataset: " + data_filename)
        if INTERACTIVE_PLOT:
            plt.show()
        else:
            plt.savefig(f"{output_base}-sklearn.png")
    if XGBOOST_PLOT:
        # note: based on https://stackoverflow.com/questions/46943314/xgboost-plot-importance-doesnt-show-feature-names
        ## TODO:plot_importance(model).set_yticklabels(feature_labels)
        ## TEST: model.feature_names = feature_labels
        ## model.feature_types = None
        ## TEST: model.feature_names = labels
        ## model.get_booster().feature_names = labels
        ## plot_importance(model.get_booster())
        ## SO-SO: plot_importance(model)
        ax = plot_importance(model)
        ## OK???: ax.set_yticklabels(feature_labels)
        debug.trace_object(6, ax, "plot axis")
        plt.figtext(0, 0, "dataset: " + data_filename)
        if INTERACTIVE_PLOT:
            plt.show()
        else:
            plt.savefig(f"{output_base}-xgb.png")

    # make predictions for test data and evaluate
    predictions = model.predict(X_test)
    accuracy = accuracy_score(y_test, predictions)
    print("Accuracy: %.2f%%" % (accuracy * 100.0))

    # Fit model using each importance as a threshold
    ## TODO: debug.trace_expr(4, zip(labels, model.feature_importances_))
    debug.trace_expr(4, labels, model.feature_importances_, delim="\n")
    thresholds = np.sort(model.feature_importances_)
    for thresh in thresholds:
        # select features using threshold
        selection = SelectFromModel(model, threshold=thresh, prefit=True)
        select_X_train = selection.transform(X_train)

        # train model
        selection_model = XGBClassifier()
        selection_model.fit(select_X_train, y_train)
        #
        # eval model
        select_X_test = selection.transform(X_test)
        predictions = selection_model.predict(select_X_test)
        accuracy = accuracy_score(y_test, predictions)

        # summarize
        print("Thresh=%.3f, n=%d, Accuracy: %.2f%%" % (thresh, select_X_train.shape[1], accuracy*100.0))
            
#-------------------------------------------------------------------------------

if __name__ == "__main__":
    main()



#! /usr/bin/env python3
# 
# Based on following:
#   https://scikit-learn.org/stable/auto_examples/ensemble/plot_forest_importances.html.
#

"""
==========================================
Feature importances with a forest of trees
==========================================

This example shows the use of a forest of trees to evaluate the importance of
features on an artificial classification task. The blue bars are the feature
importances of the forest, along with their inter-trees variability represented
by the error bars.

As expected, the plot suggests that 3 features are informative, while the
remaining are not.

"""

# Standard modules
import time

# Installed modules
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.datasets import make_classification
from sklearn.ensemble import RandomForestClassifier
from sklearn.inspection import permutation_importance
from sklearn.model_selection import train_test_split

# Local modules
from mezcla import debug
from mezcla import system
from mezcla.system import getenv_bool, getenv_int, getenv_text, getenv_value

# Constants
TL = debug.TL
INTERACTIVE_PLOT =  getenv_bool("INTERACTIVE_PLOT", False,
                                "Show interactive plot")
DEFAULT_OUTBASE = (system.remove_extension(__file__) or "plot_forest_importance")
OUTPUT_BASE = getenv_text("OUTPUT_BASE", DEFAULT_OUTBASE,
                          "Base for output files")
# note: four different individual seeds are used in order to reproduce original example;
# but, for convenience, a single seed can be used
RANDOM_SEED = getenv_value("RANDOM_SEED", None,
                           "Global seed for convenient randomization")
SEPARATE_SEEDS = (RANDOM_SEED is None)
DATASET_SEED = getenv_int("DATASET_SEED", (0 if SEPARATE_SEEDS else None),
                          "Random seed for dataset generation via make_classification")
SPLIT_SEED = getenv_int("SPLIT_SEED", (42 if SEPARATE_SEEDS else None),
                        "Random seed for dataset split")
FOREST_SEED = getenv_int("FOREST_SEED", (0 if SEPARATE_SEEDS else None),
                         "Random seed for forest generation")
PERMUTATION_SEED = getenv_int("PERMUTATION_SEED", (42 if SEPARATE_SEEDS else None),
                              "Random seed for permutation")

#-------------------------------------------------------------------------------

def main():
    """Entry point"""
    debug.trace(TL.USUAL, f"main(): filename={system.real_path(__file__)}")

    # Optionally set global seed
    if (RANDOM_SEED is not None):
        np.random.seed(system.to_int(RANDOM_SEED))

    # %%
    # Data generation and model fitting
    # ---------------------------------
    # We generate a synthetic dataset with only 3 informative features. We will
    # explicitly not shuffle the dataset to ensure that the informative features
    # will correspond to the three first columns of X. In addition, we will split
    # our dataset into training and testing subsets.
    ## OLD
    ## from sklearn.datasets import make_classification
    ## from sklearn.model_selection import train_test_split
    
    X, y = make_classification(
        n_samples=1000,
        n_features=10,
        n_informative=3,
        n_redundant=0,
        n_repeated=0,
        n_classes=2,
        random_state=DATASET_SEED,
        shuffle=False,
    )
    debug.trace_expr(TL.VERBOSE, X, y)

    # Optionally write data to CSV file
    def write_csv_file(X, y, feature_names=None, path=None):
        """Write CSV file using X and Y, optionally using FEATURE_NAMES and PATH"""
        if feature_names is None:
            num_features = len(X[0]) if len(X) else 0
            feature_names = ([f"F{i + 1}" for i in range(num_features)] + ["class"])
        if path is None:
            path = OUTPUT_BASE + ".out.csv"
        lines = []
        lines.append(",".join(feature_names))
        for features, classification in zip(X, y):
            # note: need to use list() to avoid numpy numeric operation
            lines.append(",".join(map(str, list(features) + [classification])))
        system.write_lines(path, lines)
    #
    debug.code(TL.DETAILED, lambda: write_csv_file(X, y))
               
    X_train, X_test, y_train, y_test = train_test_split(X, y,
                                                        stratify=y,
                                                        random_state=SPLIT_SEED)
    debug.trace(4, f"split sizes: train={len(X_train)}; test={len(X_test)}")
    debug.trace_expr(TL.VERBOSE, X_train, X_test, y_train, y_test)
    
    # %%
    # A random forest classifier will be fitted to compute the feature importances.
    ## OLD: from sklearn.ensemble import RandomForestClassifier
    
    feature_names = [f"feature {i}" for i in range(X.shape[1])]
    forest = RandomForestClassifier(random_state=FOREST_SEED)
    forest.fit(X_train, y_train)
    debug.trace_object(TL.VERBOSE, forest, "forest model")
    
    # %%
    # Feature importance based on mean decrease in impurity
    # -----------------------------------------------------
    # Feature importances are provided by the fitted attribute
    # `feature_importances_` and they are computed as the mean and standard
    # deviation of accumulation of the impurity decrease within each tree.
    #
    # .. warning::
    #     Impurity-based feature importances can be misleading for **high
    #     cardinality** features (many unique values). See
    #     :ref:`permutation_importance` as an alternative below.
    ## OLD: import time
    ## OLD: import numpy as np
    
    start_time = time.time()
    importances = forest.feature_importances_
    std = np.std([tree.feature_importances_ for tree in forest.estimators_], axis=0)
    elapsed_time = time.time() - start_time
    
    print(f"Elapsed time to compute the impurity-based importances: {elapsed_time:.3f} seconds")
    
    # %%
    # Let's plot the impurity-based importance.
    ## OLD: import pandas as pd
    
    forest_importances = pd.Series(importances, index=feature_names)
    
    fig, ax = plt.subplots()
    forest_importances.plot.bar(yerr=std, ax=ax)
    ax.set_title("Feature importances using MDI")
    ax.set_ylabel("Mean decrease in impurity (MDI)")
    fig.tight_layout()
    if not INTERACTIVE_PLOT:
        plt.savefig(f"{OUTPUT_BASE}-mdi.png")
    
    # %%
    # We observe that, as expected, the three first features are found important.
    #
    # Feature importance based on feature permutation
    # -----------------------------------------------
    # Permutation feature importance overcomes limitations of the impurity-based
    # feature importance: they do not have a bias toward high-cardinality features
    # and can be computed on a left-out test set.
    ## OLD: from sklearn.inspection import permutation_importance
    
    start_time = time.time()
    result = permutation_importance(
        forest, X_test, y_test, n_repeats=10, random_state=PERMUTATION_SEED, n_jobs=2
    )
    elapsed_time = time.time() - start_time
    print(f"Elapsed time to compute the permutation-based importances: {elapsed_time:.3f} seconds")
    
    forest_importances = pd.Series(result.importances_mean, index=feature_names)
    
    # %%
    # The computation for full permutation importance is more costly. Features are
    # shuffled n times and the model refitted to estimate the importance of it.
    # Please see :ref:`permutation_importance` for more details. We can now plot
    # the importance ranking.
    
    fig, ax = plt.subplots()
    forest_importances.plot.bar(yerr=result.importances_std, ax=ax)
    ax.set_title("Feature importances using permutation on full model")
    ax.set_ylabel("Mean accuracy decrease")
    fig.tight_layout()
    if not INTERACTIVE_PLOT:
        plt.savefig(f"{OUTPUT_BASE}-perm.png")

    # Show both plots
    if INTERACTIVE_PLOT:
        plt.show()

    # %%
    # The same features are detected as most important using both methods. Although
    # the relative importances vary. As seen on the plots, MDI is less likely than
    # permutation importance to fully omit a feature.

    return

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    main()




#! /usr/bin/env python3
#
# Issues query via Bing Search API

# Notes:
# - The results can be cached to a local file to avoid using up search quota when
#   debugging with the same query. Multiple cached results are stored under temp dir.
# - Requires a Bing Search API key, which can be obtained via Windows Azure Marketplace:
#      https://azure.microsoft.com
# - Currently only 1000 queries per month are allowed without fee. (Previously it was 5k!)
# - Based originally on following tips:
#      https://stackoverflow.com/questions/27606478/search-bing-via-azure-api-using-python
# - Updgraded for Azure Cognitive Services:
#      https://docs.microsoft.com/en-us/azure/cognitive-services/bing-web-search/quickstarts/python
# - For the latest API as of Spring 2020, see
#      https://docs.microsoft.com/en-us/rest/api/cognitiveservices-bingsearch/bing-web-api-v7-reference
# - For operators that can be used, see
#      https://support.microsoft.com/en-us/topic/advanced-search-keywords-ea595928-5d63-4a0b-9c6b-0b769865e78a
# - New source for key and other info:
#      https://portal.azure.com/#view/Microsoft_Bing_Api
#      https://learn.microsoft.com/en-us/bing/search-apis/bing-web-search/quickstarts/rest/python
#
# TODO:
# - *** Have option to just show hit count. ***
# - Have option to just show results for misspelling (not Bing's replacement).
# - Have option to specify user agent.
# - Have option to download document for each result.
#

"""Issues queries via Bing Search API (using MS Azure Cognitive Services)"""

# Standard packages
## TEST: import base64
import json
import sys
from six.moves.urllib_parse import quote as quote_url       # pylint: disable=import-error
from six.moves.urllib.request import Request, build_opener  # pylint: disable=import-error
import tempfile

# Local packages
from mezcla import tpo_common as tpo
# TODO: import xml.dom.minidom
from mezcla import debug
from mezcla import glue_helpers as gh
from mezcla import system

BING_KEY = (tpo.getenv_value("BING_KEY", None,
                             "API key (via Microsoft Azure)") or "")
BING_BASE_URL = tpo.getenv_text("BING_BASE_URL",
                                "https://api.bing.microsoft.com/v7.0/")
SEARCH = "search"
NEWS = "news/search"
IMAGES = "images"
VALID_SEARCH_TYPES = {SEARCH, IMAGES, "videos", NEWS}
DEFAULT_SEARCH_TYPE = SEARCH
SEARCH_TYPE = system.getenv_text("SEARCH_TYPE", SEARCH)

## OLD:
## DEFAULT_TEMP_FILE = tempfile.NamedTemporaryFile().name
## TEMP_FILE = system.getenv_text("TEMP_FILE", DEFAULT_TEMP_FILE)
## DEFAULT_TEMP_DIR = tempfile.gettempdir()
## TEMP_DIR = system.getenv_text("TEMP", DEFAULT_TEMP_DIR)
TEMP_FILE = gh.TEMP_FILE
TEMP_DIR = system.TEMP_DIR

USE_CACHE = system.getenv_bool("USE_CACHE", False)

#...............................................................................

def bing_search(query, key=None, use_json=None, search_type=None, topn=None, non_phrasal=False):
    """Issue QUERY bing using API KEY returning result, optionally using JSON format by default or with alternative SEARCH_TYPE (e.g., image) or limited to TOPN results. The query is quoted unless NON_PHRASAL.
    Note: SEARCH_TYPE in {search, images, videos, news, SpellCheck}."""
    ## TODO: see if old search types RelatedSearch and Composite still supported
    tpo.debug_print("bing_search(%s, %s, %s, %s, %s)" % (query, key, use_json, search_type, topn), 4)
    if search_type is None:
        search_type = SEARCH_TYPE
    if use_json is None:
        use_json = True
    debug.assertion(search_type in VALID_SEARCH_TYPES)
    if key is None:
        key = BING_KEY
    full_query = query if non_phrasal else ("'" + query + "'")
    query_spec = quote_url(full_query)

    # Create credential for authentication
    # TODO: make user_agent an option, along with ip_address
    user_agent = "Nonesuch/1.0 (python)"
    ## TEST:
    ## try:
    ##     colon_key_bytes = (":" + key).encode("utf-8")
    ##     encoded_colon_key = base64.b64encode(colon_key_bytes)
    ## except(TypeError, ValueError):
    ##     system.print_stderr("Exception in bing_search: {exc}", exc=sys.exc_info())
    ##     encoded_colon_key = ":???"
    ## auth = "Basic %s" % encoded_colon_key

    # Format search URL with optional format and top-n specification
    # TODO2: drop unused format_spec
    debug.assertion(use_json)
    format_spec = ""
    topn_spec = ("&count=%d" % topn) if topn else ""
    tpo.debug_print("format_spec=%s topn_spec=%s" % (format_spec, topn_spec), 5)
    sources_spec = ""
    # HACK: if multiple types specified, comvert into Composite search
    if " " in search_type:
        sources = quote_url("'" + "+".join(search_type.split(" ")) + "'")
        sources = sources.replace("SpellingSuggestion", "spell").lower()
        sources_spec = "&Sources=" + sources
        search_type = "Composite"
    url_params = search_type + "?q=" + query_spec + sources_spec + topn_spec + format_spec
    url = BING_BASE_URL + "/" + url_params

    # Check cache if available
    response_data = None
    cache_file = system.form_path(TEMP_DIR, "_bs-" + url_params)
    if USE_CACHE and system.non_empty_file(cache_file):
        debug.trace_fmt(4, "Using cached results: {cf}", cf=cache_file)
        response_data = system.read_file(cache_file)
    
    # Download data from URL (unless cached).
    # Note: Also caches result
    if not response_data:
        tpo.debug_print("Accessing URL %s" % url, 3)
        request = Request(url)
        ## OLD: request.add_header("Authorization", auth)
        request.add_header("Ocp-Apim-Subscription-Key", key)
        request.add_header("User-Agent", user_agent)
        tpo.debug_print("Headers: %s" % request.header_items(), 5)
        request_opener = build_opener()
        response = request_opener.open(request) 
        response_data = response.read()
        if USE_CACHE:
            system.write_file(cache_file, response_data)

    # Format result
    tpo.debug_print("Response: %s" % response_data, 5)
    if use_json:
        json_result = json.loads(response_data)
        if debug.verbose_debugging():
            ## TODO: system.write_file(TEMP_FILE, str(json_result))
            ## TEST: hash_text = "\n".join([(str(k) + ": " + str(h)) for (k, h) in json_result.items()])
            ## system.write_file(TEMP_FILE, str(json_result))
            system.write_file(TEMP_FILE, system.to_string(response_data))
        ## TODO: result_list = json_result["webPages"]["value"]
        result_list = json_result
    else:
        result_list = response_data
        ## TODO: xml_result = xml.dom.minidom.parseString(response_data)
        ## TODO: result_list = xml_result.toprettyxml()
    ## TODO2:" result_list => result_hash
    debug.trace_expr(5, result_list, max_len=4096)
    return result_list

#...............................................................................

def main():
    """Entry point for script"""
    search_type = SEARCH
    use_json = True
    
    # Check command-line arguments (TODO: convert to main.py)
    i = 1
    show_usage = (i >= len(sys.argv))
    while (i < len(sys.argv) and (sys.argv[i][0] == "-")):
        if (sys.argv[i] == "--json"):
            use_json = True
        elif (sys.argv[i] == "--xml"):
            use_json = False
        elif (sys.argv[i] == "--image"):
            search_type = IMAGES
        elif (sys.argv[i] == "--type"):
            i += 1
            search_type = sys.argv[i]
        else:
            show_usage = True
        i += 1
    if (show_usage):
        print("Usage: %s [--json | --xml] [--image] [--type label] query_word ..." % sys.argv[0])
        print("")
        print("Notes:")
        print("- Set BING_KEY to key obtained via Microsoft Azure; see following:")
        print("  https://learn.microsoft.com/en-us/azure/cognitive-services/bing-web-search")
        print("- Types: Web, Image, Video, News, SpellingSuggestion, RelatedSearch")
        print("- For API details, see following:")
        print("  https://docs.microsoft.com/en-us/rest/api/cognitiveservices-bingsearch/bing-web-api-v7-reference")
        print("- The resuslt is XML file with <entry> tags for blurb info")
        sys.exit()
    debug.assertion(BING_KEY)

    # Issue query and print result
    query = " ".join(sys.argv[i:])
    print(bing_search(query, use_json=use_json, search_type=search_type))
    return


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""The basic unit of TF-IDF, the keyword.

This class allows you to have stemmed keywords, but still see the original text.

Warning: Note that this is low-level bookkeeping code, which is reflected in
the following example, which incorporates hard-coded stemming. See preprocess.py 
for usage with an actual stemmer.

Example:
    >>> import mezcla.tfidf.dockeyword as mtk
    >>> import mezcla.tfidf.document as mtd
    >>> text = 'dog dogs'
    >>> dog = mtk.DocKeyword('dog', mtd.Document(text), start=0, end=3)
    >>> dogs = mtk.DocKeyword('dog', mtd.Document(text), start=4, end=8)
    >>> dog.text == dogs.text
    True
    >>> dog.original_texts == dogs.original_texts
    False
    >>> str(dog)
    "Stem:dog, Instances:['dog'], Count:1"
    >>> str(dogs)
    "Stem:dog, Instances:['dogs'], Count:1"
"""

# Standard modules
from __future__ import absolute_import
from collections import namedtuple

# Local modules
from mezcla.tfidf.config import BASE_DEBUG_LEVEL as BDL
from mezcla import debug
from mezcla import system

# Constants
Location = namedtuple('Location', ['document', 'start', 'end'])
ADD_DUMMY_LOCATION = system.getenv_boolean(
    ## TODO2: drop and just use SKIP_LOCATION
    "ADD_DUMMY_LOCATION", False,
    desc="Add dummy location for debugging purposes")
SKIP_DOC_LOCATION = system.getenv_boolean(
    "SKIP_DOC_LOCATION", False,
    desc="Skip location tracking")


class DocKeyword(object):
    """Class for maintaining stemmed term and original"""
    # Note: debug tracing commented out to cut down on overhead
    
    def __init__(self, text, document=None, start=None, end=None,
                 skip_location=None, add_dummy_location=None):
        self.locations = set()
        self.text = text
        self.count = 1
        if skip_location is None:
            skip_location = SKIP_DOC_LOCATION
        self.skip_location = skip_location
        if add_dummy_location is None:
            add_dummy_location = ADD_DUMMY_LOCATION
        self.add_dummy_location = add_dummy_location
        debug.assertion(not (self.skip_location and add_dummy_location))
        if self.add_dummy_location:
            if (start is None):
                start = -1
            if (end is None):
                end = -1
        if (start is not None) and (end is not None):
            self.locations.add(Location(document, start, end))
        ## DEBUG: debug.trace_object(BDL + 4, self, "DocKeyword instance", show_all=False)

    def update_locations(self, locations):
        """Add LOCATIONS to other locations"""
        self.locations = self.locations.union(locations)

    def __add__(self, other):
        assert self.text == other.text
        out = DocKeyword(self.text)
        out.locations = self.locations
        out.update_locations(other.locations)
        return out

    def __ladd__(self, other):
        if other == 0:
            return self
        return self.__add__(other)

    def __len__(self):
        length = len(self.locations) if not self.skip_location else self.count
        return length

    @property
    def original_texts(self):
        """Return list of original texts"""
        out = []
        if self.skip_location:
            ## TODO3: just show warning once
            debug.trace(BDL - 1, "FYI: original_texts disabled with skip_location")
        for loc in self.locations:
            if loc.document:
                text = loc.document.text[loc.start: loc.end]
            else:
                text = ''
            out.append(text)
        ## DEBUG: debug.trace_expr(BDL + 3, out)
        return list(set(out))

    def get_first_text(self):
        """Return the first original text."""
        if self.skip_location:
            ## TODO3: just show warning once
            debug.trace(BDL - 1, "FYI: get_first_text disabled with skip_location")
        loc = next(iter(self.locations))
        ## DEBUG: debug.trace_expr(BDL + 3, loc.document.text)
        return loc.document.text[loc.start: loc.end]

    def __str__(self):
        return ('Stem:%s, Instances:%s, Count:%d, Len:%d, #Locs:%d'
                % (self.text, str(self.original_texts), self.count, len(self.locations), len(self)))

#-------------------------------------------------------------------------------
    
if __name__ == '__main__':
    system.print_stderr(f"Warning: {__file__} is not intended to be run standalone")
    monkey_keyword = DocKeyword("monkeys")
    print(f"{monkey_keyword=}")
    print(f"{str(monkey_keyword)=}")
    debug.trace_object(BDL - 2, monkey_keyword, "monkey_keyword")



#! /usr/bin/env python3
#
# Uses BERT or ALBERT for multiple-class text classification, based on following blog:
#    https://towardsdatascience.com/beginners-guide-to-bert-for-multi-classification-task-92f5445c2d7c
#
# Notes:
# - Setup instructions:
#   -- sudo pip3 install virtualenv
#   -- virtualenv bertenv
#   -- python3 -m venv bertenv
#   -- source bertenv/bin/activate
#   -- tensorflow >= 1.11.0   # CPU Version of TensorFlow.
#   -- tensorflow-gpu  >= 1.11.0  # GPU version of TensorFlow
# - Format for input files
# - Currently, the input files must use the basename "train", "dev" and optionally "test". CSV import must
#   use the extension .csv, and likewise .tsv for Tab-separated values (TSV).
# - BERT itself requires following format in tab-separated value format (tsv):
#     1. guid: Unique id for the example.
#     2. label: String data. The label of the example. This should be specified for train and evaluate examples, but not for test examples.
#     3. text_b: (Optional) string data. The untokenized text of the second sequence. Only must be specified for sequence pair tasks.
#     4. text_a: String data. The untokenized text of the first sequence. For single sequence tasks, only this sequence must be specified.
#
# - Sample preprocessing
#     - Add unique ID to training and dev-test examples, and make dummy third column.
#       -- The first perl command adds the dummy - column, and the second replaces the
#       leading spaces prior with prefix for line number added by cat. The prefix is
#       based on the size of the training file to ensure ID is unique for dev-test.
#       TODO: drop prefix support as BERT uses file type (e.g., train/dev/test).
#       $: {
#          function convert_to_bert() {
#             local file="$1"
#             local prefix="$2"
#             perl -pe 's/\t/\t-\t/;' "$file" | cat -n | perl -pe "s/^\s*/$prefix/;"
#          }
#          #
#          base=pruned-random-simplewiki-20170720-pages-articles.texts.data
#          convert_to_bert $base.train "" > train.tsv
#          prefix=$(wc -l < train.tsv)
#          convert_to_bert $base.test "$prefix" > dev.tsv
#        }
#     - Add header to dummy testing data
#       echo $'guid\ttext' > test.tsv
#       TODO: create test data by removing cases from other two files
#     - Create category file
#       cut -f2 train.tsv dev.tsv | sort --unique > cats.txt
# - Sample invocation:
#      init-miniconda3
#      source activate albert_py3
#      #
#      export DATA_DIR=$PWD;
#      export MODEL_DIR=/usr/local/misc/data/deep-learning/albert/albert_base;
#      export OUTPUT_DIR=$DATA_DIR/output;
#      export LOWER_CASE=1
#      export TASK_NAME=multi
#      export CLASSIFIER_INVOCATION="$PYTHON -m run_albert_classifier"
#   
#      # Run albert classifier indirection through wrapper script 
#      _base_=bert_multi_classification
#      # TODO: ???
#      let _RL_++; _log_="$DATA_DIR/$_base_-$(todays-date).log$_RL_"; DEBUG_LEVEL=4 USE_TSV_INPUT=1 $PYTHON -m $_base_ > "$_log_" 2>&1; less-tail "$_log_"   
#
# TODO:
# - * Summary evaluation over test set!
# - Use Main class to add in support for argument parsing
# - Generalize to handling other data files.
# - Add options for BERT model and other parameters:
#   ex: train_batch_size, learning_rate, max_seq_length, and num_train_epochs.
# - Convert into calling module for run_classifier.py directly (i.e., without using shell).
# - Rework sample invocation so that environment settings are temporary.
# - Add option for using common defaults (e.g., data in current dir, lowercase, base model, output in ./output, multi-category, using run_albert_classifier.py).
# - Create random split of single input file into train.tsv, dev.tsv and optionally test.tsv (or .csv's).
#

"""Invokes BERT or ALBERT for multiple-class text classification"""

# Standard packages
import csv
import re
import sys

# Installed packages
# Note: imports tensorflow before others due to its stupid exception quirks with sklearn
# See https://github.com/tensorflow/tensorflow/issues/2034 for a similar problem with numpy.
import tensorflow as tf
import pandas as pd
from sklearn.model_selection import train_test_split

# Local packages
from mezcla import debug
from mezcla import system
from mezcla import glue_helpers as gh
from mezcla.text_utils import version_to_number as version_as_float

# Note: the defaults are unintuitive, but they match the blog article.
MODEL_DIR = system.normalize_dir(system.getenv_text("MODEL_DIR", "./model"))
## TODO: BERT_DIR = system.normalize_dir(system.getenv_text("BERT_DIR", MODEL_DIR))
DATA_DIR = system.normalize_dir(system.getenv_text("DATA_DIR", "./dataset"))
OUTPUT_DIR = system.normalize_dir(system.getenv_text("OUTPUT_DIR", "./bert_output"))
TASK_NAME = system.getenv_text("TASK_NAME", "cola")
USE_TSV_INPUT = system.getenv_bool("USE_TSV_INPUT", False)
CLASSIFIER_INVOCATION = system.getenv_text("CLASSIFIER_INVOCATION", "run_bert_classifier.py")
# Note: USE_TSV_INPUT implies you use a tab-separated format properly formatted for BERT,
# so set BERT_FORMATTED False (e.g., USE_TSV_INPUT=1 BERT_FORMATTED=0 bert_multi_classification.py ...)
BERT_FORMATTED = system.getenv_bool("BERT_FORMATTED", USE_TSV_INPUT)
LOWER_CASE = system.getenv_bool("LOWER_CASE", False)
USE_ALBERT_DEFAULT = ("albert" in CLASSIFIER_INVOCATION)
USE_ALBERT = system.getenv_bool("USE_ALBERT", USE_ALBERT_DEFAULT)
BERT_NAME = "bert" if (not USE_ALBERT) else "albert"
CONFIG_FILE_DEFAULT = system.form_path(MODEL_DIR, "{b}_config.json".format(b=BERT_NAME))
BERT_CONFIG_FILE = system.getenv_text("BERT_CONFIG_FILE", CONFIG_FILE_DEFAULT)
SAVE_CHECKPOINTS_STEPS = system.getenv_int("SAVE_CHECKPOINTS_STEPS", 10000,
                                           "Number of tensorflow steps before checkpoints are written")

# Get label list and split into columns
INPUT_LABELS = system.getenv_text("INPUT_LABELS", "id, label, text")
## TODO: INPUT_LABEL_LIST = re.split(r"\s, *", INPUT_LABELS)
# Labels to use for ID, CLASS_VALUE, and TEXT
INPUT_LABEL_LIST = re.split(r",", INPUT_LABELS)
debug.assertion(len(INPUT_LABEL_LIST) == 3)
ID_COL = INPUT_LABEL_LIST[0]
LABEL_COL = INPUT_LABEL_LIST[1]
TEXT_COL = INPUT_LABEL_LIST[2]

#-------------------------------------------------------------------------------
# Helper function(s) specific to BERT

def ensure_bert_data_frame(data_frame, is_test=False):
    """Ensures data frame is in BERT format from input DATA_FRAME, using dummy values for alpha
    column.
    Notes:
    - See comments in blog mentioned in header above.
    - Uses global constant BERT_FORMATTED."""
    # TODO: add parameter mapping input column names to ones assumed here (i.e., id, label, & text)
    debug.trace_fmt(5, "ensure_bert_data_frame({df})", df=data_frame)
    df_bert = None
    if BERT_FORMATTED:
        df_bert = data_frame
    else:
        try:
            first_sentence = data_frame[TEXT_COL][0]
            debug.trace_fmt(5, "First sentence: {s}", s=first_sentence)
            # Ignore header column if given (TODO, add parameter to make this optional)
            # Note: This assumes single word tokens without punctuation can't be a sentence.
            if re.search(r"^\w+$", first_sentence):
                debug.trace(3, "Removing presumed header row of (test) data frame")
                debug.assertion(is_test)
                data_frame = data_frame.drop(data_frame.index[0])
                
            data_hash = {'guid': data_frame[ID_COL],
                         'alpha': (['-'] * data_frame.shape[0]),
                         'text': data_frame[TEXT_COL]}
            if not is_test:
                data_hash['label'] = data_frame[LABEL_COL]
                
            df_bert = pd.DataFrame(data_hash)
        except:
            debug.raise_exception(5)
            system.print_stderr("Exception converting data frame to BERT format: {exc}",
                                exc=sys.exc_info())
    debug.trace_fmt(4, "ensure_bert_data_frame(_) => {r}", r=df_bert)
    return df_bert

#--------------------------------------------------------------------------------

# Note: BERT works well with tensorflow version 1.15
debug.assertion(version_as_float("1.11") <= version_as_float(tf.version.VERSION) < version_as_float("2.0"))

#-------------------------------------------------------------------------------

## TODO: changes to BERT run_classifier.py code for different label sets
## NOTE: not needed if running one of GLUE tasks with support built into classifier (e.g., CoLA)
## def get_labels(self):
##     return ["0", "1"]
## def get_labels(self):
##    return ["0", "1", "2", "3", "4"]
## def get_labels(self):
##    return ["POSITIVE", "NEGATIVE"]

def main():
    """Entry point for script"""
    ## sudo apt-get install python3-pip
    ## a550d    1    a    To clarify, I didn't delete these pages.kcd12    0    a    Dear god this site is horrible.7379b    1    a    I think this is not appropriate.cccfd    2    a    The title is fine as it is.
    ## guid     textcasd4    I am not going to buy this useless stuff.3ndf9    I wanna be the very best, like no one ever was
    ## pip install pandas
    ## pip install sklearn
    ## id,text,labelsadcc,This is not what I want.,1cj1ne,He seriously have no idea what it is all about,0123nj,I don't think that we have any right to judge others,2
    in_separator = ","
    in_ext = ".csv"
    in_quoting = csv.QUOTE_MINIMAL
    if USE_TSV_INPUT:
        in_separator = "\t"
        in_ext = ".tsv"
        in_quoting = csv.QUOTE_NONE
    debug.trace_expr(5, in_separator, in_ext, in_quoting)
    df_train = pd.read_csv(gh.form_path(DATA_DIR, "train" + in_ext), sep=in_separator, names=INPUT_LABEL_LIST, quoting=in_quoting)
    df_bert_train = ensure_bert_data_frame(df_train)

    ## BAD: df_bert_train.to_csv(gh.form_path(DATA_DIR, 'train.tsv'), sep='\t', index=False, header=False)
    ## TODO: df_train.to_csv(gh.form_path(DATA_DIR, 'train.tsv'), sep='\t', index=False, header=False)

    # Read source data from csv file
    test_columns = [INPUT_LABEL_LIST[0], INPUT_LABEL_LIST[2]]
    df_test = pd.read_csv(gh.form_path(DATA_DIR, "test" + in_ext), sep=in_separator, names=test_columns, quoting=in_quoting)
    df_bert_test = ensure_bert_data_frame(df_test, is_test=True)

    ## TODO: alternative version
    ## #create a new dataframe for train, dev data
    ## df_bert = pd.DataFrame({'guid': df_train['id'],
    ##                         'label': df_train['label'],
    ##                         'alpha': ['a']*df_train.shape[0],
    ##                         'text': df_train['text']})

    # Optionally split into test, dev
    # NOTE: skipped if no dev.tsv file exists
    dev_file = gh.form_path(DATA_DIR, "dev" + in_ext)
    if system.file_exists(dev_file):
        df_dev = pd.read_csv(dev_file, sep=in_separator, names=INPUT_LABEL_LIST, quoting=in_quoting)
        df_bert_dev = ensure_bert_data_frame(df_dev)
    else:
        df_bert_train, df_bert_dev = train_test_split(df_bert_train, test_size=0.01)  
        
    ## ALT: create new dataframe for test data
    ## df_bert_test = ensure_bert_data_frame(df_test)
    ## pd.DataFrame({'guid': df_test['id'],
    ##               'text': df_test['text']})

    # Output tsv file, no header for train and dev
    if not USE_TSV_INPUT:
        df_bert_train.to_csv(gh.form_path(OUTPUT_DIR, 'train.tsv'), sep='\t', index=False, header=False)
        df_bert_dev.to_csv(gh.form_path(OUTPUT_DIR, 'dev.tsv'), sep='\t', index=False, header=False)
        df_bert_test.to_csv(gh.form_path(OUTPUT_DIR, 'test.tsv'), sep='\t', index=False, header=True)

    # Sanity checks
    debug.assertion(system.file_exists("cats.txt"))
    #
    ## TODO: work example error from run_classifier.py customization into assertion
    ## label_id = label_map[example.label]
    ## KeyError: '2'`

    ## TODO: Run NVIDIA CUDA utility and make sure capable of running TensorFlow w/ GPU's.
    ## Also, warn if graphics memory is too low.
    ## nvidia-smi
    system.setenv("BERT_BASE_DIR", MODEL_DIR)
    ##CUDA_VISIBLE_DEVICES=0
    ## python script.py
    print("Make sure your GPU Processor has sufficient memory, besides adequate number of units")
    print(gh.run("nvidia-smi"))
    # Make sure tensorflow doesn't grab all the GPU memory
    if system.getenv("TF_FORCE_GPU_ALLOW_GROWTH") is None:
        system.setenv("TF_FORCE_GPU_ALLOW_GROWTH", "true")
    # note: 0 is the order, not the total number
    if system.getenv("CUDA_VISIBLE_DEVICES") is None:
        system.setenv("CUDA_VISIBLE_DEVICES", "0")
    if system.getenv("NVIDIA_VISIBLE_DEVICES") is None:
        system.setenv("NVIDIA_VISIBLE_DEVICES", "0")
    # TODO: use run [?] and due sanity checks on output; u[???}
    is_lower_case = system.to_string(LOWER_CASE).lower()
    # TODO: use form_path here and below
    bert_proper_args = ("--vocab_file={md}/vocab.txt".format(md=MODEL_DIR) if (not USE_ALBERT) else "--spm_model_file={md}/albert.model".format(md=MODEL_DIR))
    # TODO: add more env params: MAX_SEQ_LENGTH, TRAIN_BATCH_SIZE, LEARNING_RATE, NUM_TRAIN_EPOCHS
    # TODO: specify checkpoint separately: {md}/{bn}_model.ckpt => MODEL_CHECKPOINT
    ## OLD: gh.issue("{ci} {bpa} --task_name={t} --do_train=true --do_eval=true --do_predict=true --data_dir={dd}  --{bn}_config_file={bcf} --init_checkpoint={md}/{bn}_model.ckpt --max_seq_length=64 --train_batch_size=2 --learning_rate=2e-5 --num_train_epochs=3.0 --output_dir={od} --do_lower_case={lc} --save_checkpoints_steps={scs}", ci=CLASSIFIER_INVOCATION, bpa=bert_proper_args, t=TASK_NAME, dd=DATA_DIR, md=MODEL_DIR, od=OUTPUT_DIR, lc=is_lower_case, bn=BERT_NAME, bcf=BERT_CONFIG_FILE, scs=SAVE_CHECKPOINTS_STEPS)
    ## Uses: defaults for following: train_batch_size (32); learning_rate (5e5); max_seq_length (128); num_train_epochs (3)
    ## See run_bert_classifier.py for other options (e.g., --iterations_per_loop).
    gh.issue("{ci} {bpa} --task_name={t} --do_train=true --do_eval=true --do_predict=true --data_dir={dd}  --{bn}_config_file={bcf} --init_checkpoint={md}/{bn}_model.ckpt --output_dir={od} --do_lower_case={lc} --save_checkpoints_steps={scs}", ci=CLASSIFIER_INVOCATION, bpa=bert_proper_args, t=TASK_NAME, dd=DATA_DIR, md=MODEL_DIR, od=OUTPUT_DIR, lc=is_lower_case, bn=BERT_NAME, bcf=BERT_CONFIG_FILE, scs=SAVE_CHECKPOINTS_STEPS)
    ## sample output:
    ## eval_accuracy = 0.96741855 eval_loss = 0.17597112 global_step = 236962 loss = 0.17553209
    ## checkpoint file: model_checkpoint_path: "model.ckpt-236962" all_model_checkpoint_paths: "model.ckpt-198000" all_model_checkpoint_paths: "model.ckpt-208000"  all_model_checkpoint_paths: "model.ckpt-218000"  all_model_checkpoint_paths: "model.ckpt-228000"  all_model_checkpoint_paths: "model.ckpt-236962"
    
    ## alternative run
    ## CUDA_VISIBLE_DEVICES=0 python run_classifier.py --task_name=cola --do_predict=true --data_dir=./dataset --vocab_file=./model/vocab.txt --bert_config_file=./model/bert_config.json --init_checkpoint=./bert_output/model.ckpt-236962 --max_seq_length=64 --output_dir=./bert_output/
    ## 1.4509245e-05 1.2467547e-05 0.999946361.4016414e-05 0.99992466 1.5453812e-051.1929651e-05 0.99995375 6.324972e-063.1922486e-05 0.9999423 5.038059e-061.9996814e-05 0.99989235 7.255715e-064.146e-05 0.9999349 5.270801e-06
    ## alternative input
    ## # read the original test data for the text and id
    ## df_test = pd.read_csv(gh.form_path(OUTPUT_DIR, 'test.tsv'), sep='\t')
    ## # read the results data for the probabilities
    ## df_result = pd.read_csv('bert_output/test_results.tsv', sep='\t', header=None)
    ## # create a new dataframe
    ## df_map_result = pd.DataFrame({'guid': df_test['guid'],
    ##                               'text': df_test['text'],
    ##                               'label': df_result.idxmax(axis=1)})
    ## # view sample rows of the newly created dataframe
    ## df_map_result.sample(10)
    return

#------------------------------------------------------------------------

if __name__ == '__main__':
    main()
